/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _objects_h_
#define _objects_h_

#include "Strings.h"

// CLASS GROUP: Basic Object Classes

/*---------------------------------------------------------------------
CLASS
     Enumerable

     A fundamental superclass for all objects that are enumerable.

OVERVIEW
     Provides an abstract class from which enuerable objects can
     be derived. Any class that needs to be contained within an 
     (indexed) slist, a Map, or any class derived from them, must be
     publicly derived from (i.e. must extend) the Enumerable class.

DESCRIPTION
     The basic features of an Enumerable object are that it has uniqueness,
     and an ordinal value. Hence it can thus be distinguished from other
     similar objects and arranged in order.
     When it is created, an Enumerable calls the virtual function newObjectID()
     which returns a new unique integer identifier for this object.
     When the Object class is extended by a sub-class, the sub
     class must redefine newObjectID() so that identifiers are allocated
     from within a sub-range specific to that sub-class.

----------------------------------------------------------------------*/

class Enumerable {

public:

  Enumerable () 
  {
    if (! (id.ival = newObjectID()) )
      id.name = newObjectName() ;
  }

  Enumerable (char* n)
  {
    id.name = n ;
  }

  Enumerable (int i)
  {
    id.ival = i ;
  }

  virtual int newObjectID () { return 0 ; }
  virtual char* newObjectName () { return 0 ; }
  
  int index () { return id.ival ; }
  char* ident () { return id.name ; }

  void setIndex (int i) { id.ival = i ; }
  void setName (char* s) { id.name = s ; }

private:
  
  union {
    int ival ;
    char* name ;
  } id ;

};

#endif // _objects_h_
