/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

#if ! defined _statements_h_
#define _statements_h_

#include "Lists.h"
#include "Defs.h"
#include "Tags.h"
#include "Labels.h"
#include "Sets.h"
#include "Unparse.h"

/*--------------------------------------------------------------

CLASS
    Statement

    Parse tree object which represents a single statement

DESCRIPTION

 This class implements nodes representing single statements
 within a parse tree. It is derived publicly from the Tag
 class - thus tagging each Statement with the tag of the
 statement category it represents.

 Each Statement object contains pointers to the elements of a
 statement, organised as a group of type unions. Thus, each
 Statement category has a particular member within the union
 to store the parse tree entities relevant to that particular
 grammatical construct.

----------------------------------------------------------------*/

class Statement : public Tag {

friend class FortranParser ;
friend class ParseContext ;
friend class Universe<Statement> ;

private:
  int	id ;			// unique identification tag

public:

  PTR_STMT thread ;             // For Sage++/MARS compatibility only
  int g_line ;                  // global line number
  int l_line ;	                // local line number
  PTR_CNTXT	context ;	// ptr to parent context
  PTR_STMT	control_parent; // ptr to control parent
  PTR_SCOPE	scope ;         // ptr to scoping unit for this statement
  int           idx ;           // indicates section extension
  PTR_TRANS     trans ;         // link to transformation record

  union {
    PTR_FTN_LABEL ftn ;         // ptr to Fortran-style label (int)
    PTR_CC_LABEL  cc ;          // ptr to C/C++ style label (identifier)
  } label ;

  union {
    
    struct {
      PTR_STMT  bf_ptr1;	// used by the parser
      PTR_CMNT  cmnt_ptr;	// ptr to comments
      
      PTR_SYMB  symbol;		// ptr to statement's symbol
      
      PTR_EXPR  e1;             // an L-value expr tree
      PTR_EXPR  e2;             // an R-value expr tree
      PTR_EXPR  e3;             // a spare expr tree (see below)
      
      PTR_FTN_LABEL lbl_ptr;	// a secondary label
      
      PTR_BLOCK  bl_ptr1;       // list of control dep sub-nodes
      PTR_BLOCK  bl_ptr2;       // another such list (for if stmt)
    } 			fields ;             
    
    struct global {
      PTR_STMT  proc_list;	// list of procedures in this file
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  list;		// list of global consts and types
      
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      PTR_EXPR  null_4;
      
      PTR_FTN_LABEL null_5;
      
      PTR_BLOCK  control;        // used for list of procedures
      PTR_BLOCK  null_6;
    }			global ;
    
    struct program {
      PTR_STMT  next_prog ;
      PTR_CMNT  cmnt_ptr ;
      
      PTR_SYMB  prog_symb ;
      
      PTR_EXPR  null_1 ;
      PTR_EXPR  null_2 ;
      PTR_EXPR  null_3 ;
      
      LabelTable* labels ;
      
      PTR_BLOCK  control ;
      PTR_BLOCK  format_group ;
    }			program ;
    
    struct {
      PTR_STMT  next_proc;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  proc_symb;
      
      PTR_EXPR  null_1;
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      
      Universe<FtnLabel>* labels ;
      
      PTR_BLOCK  control;
      PTR_BLOCK  format_group;
    } 			procedure ;
    
    struct {
      PTR_STMT  next_func;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  func_symb;
      
      PTR_EXPR  ftype;
      PTR_EXPR  null_1;
      PTR_EXPR  null_2;
      
      LabelTable* labels ;
      
      PTR_BLOCK  control;
      PTR_BLOCK  format_group;
    }			function ;
    
    struct {
      PTR_STMT  next_bif;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      PTR_EXPR  null_4;
      
      PTR_FTN_LABEL null_5;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_6;
    }			basic_block;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      PTR_EXPR  null_4;
      
      PTR_FTN_LABEL null_5;
      
      PTR_BLOCK  null_6;
      PTR_BLOCK  null_7;
    }			 control_end;
    
    struct {
      PTR_STMT  true_branch;	// ptr to Fortran "then" statement
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  condition;	// conditional expression
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      
      PTR_FTN_LABEL null_4;
      
      PTR_STMT  then_block;	// stmt-list of "then" block
      PTR_STMT  else_block;	// stmt-list of "else" block
    }			if_node ;
    
    struct {
      PTR_STMT  true_branch;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  condition;
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      
      PTR_FTN_LABEL null_4;
      
      PTR_BLOCK  control_true;
      PTR_BLOCK  control_false;
    }                           where_node;
    
    struct {
      PTR_STMT  loop_end;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      PTR_EXPR  null_4;
      
      PTR_FTN_LABEL null_5;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_6;
    }                           loop_node;
    
    struct {
      PTR_STMT  for_end;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  control_var;
      
      PTR_EXPR  range;
      PTR_EXPR  increment;
      PTR_EXPR  where_cond;
      
      PTR_FTN_LABEL doend;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_1;
    }                           for_node;
    
    struct {
      PTR_STMT  forall_end;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  control_var;
      
      PTR_EXPR  range;
      PTR_EXPR  increment;
      PTR_EXPR  where_cond;
      
      PTR_FTN_LABEL null_1;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_2;
    }                           forall_nd;
    
    struct {
      PTR_STMT  alldo_end;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  control_var;
      
      PTR_EXPR  range;
      PTR_EXPR  increment;
      PTR_EXPR  null_0;
      
      PTR_FTN_LABEL null_1;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_2;
    }                           alldo_nd;
    
    struct {
      PTR_STMT  while_end;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  condition;
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      
      PTR_FTN_LABEL null_4;
      
      PTR_BLOCK  control;
      PTR_BLOCK  null_5;
    }                           while_node;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  condition;
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
        
      PTR_FTN_LABEL null_4;
      
      PTR_BLOCK  control_true;
      PTR_BLOCK  control_false;
    }                           exit_node;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  l_value;
      PTR_EXPR  r_value;
      PTR_EXPR  null_2;
      
      PTR_FTN_LABEL null_3;
      
      PTR_BLOCK  null_4;
      PTR_BLOCK  null_5;
    }                           assign;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  l_value;
      PTR_EXPR  r_value;
      PTR_EXPR  null_2;
      
      PTR_FTN_LABEL null_3;
      
      PTR_BLOCK  null_4;
      PTR_BLOCK  null_5;
    }                           identify;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  spec_string;
      PTR_EXPR  null_2;
      PTR_EXPR  null_3;
      
      PTR_FTN_LABEL null_4;
      
      PTR_BLOCK  null_5;
      PTR_BLOCK  null_6;
    }                           format;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  format;         // used by blaze only
      PTR_EXPR  expr_list;
      PTR_EXPR  control_list;   // used by cedar fortan only
      
      PTR_FTN_LABEL null_2;
      
      PTR_BLOCK  null_3;
      PTR_BLOCK  null_4;
    }                           write_stat;
    
    struct {
      PTR_STMT  next_stat;
      PTR_CMNT  cmnt_ptr;
      
      PTR_SYMB  null_1;
      
      PTR_EXPR  format;         // used by blaze only
      PTR_EXPR  var_list;
      PTR_EXPR  control_list;   // used by cedar fortran
      
      PTR_FTN_LABEL null_2;
      
      PTR_BLOCK  null_3;
      PTR_BLOCK  null_4;
    }                                   read_stat;

    struct {                    // C or C++ switch statements
      PTR_EXPR   switch_expr;   // the switch expression
      PTR_STMT   body;          // the switch body
      PTR_STMT   defStmt;       // the `default' statement, if present
      CaseTable* labels;        // collection of case labels
    } switch_stat;

  } node ;

public:


    ////////////////////////////////////
    //
    // Actionless constructor, used when overloading the
    // `new' operator for classes derived from Statement.
    //
  Statement () : Tag (0) { }

    ////////////////////////////////////
    //
    // Constructor used by make_member to create a identifiable 
    // Statement.
    //
  Statement (int i) : Tag (0) { id = i ; scope = 0 ; }

    ////////////////////////////////////
    //
    // After construction via the make_stmt() function
    // within the ParseContext class, the creator of a
    // Statement can then set the fields using this function.
    //
    // [in] ParseContext* the containing parse context object
    // [in] statement_tag statement category
    // [in] Symbol*       associated Symbol
    // [in] Expression*   expression 1
    // [in] Expression*   expression 2
    // [in] Expression*   expression 3
    // [in] int           local line number
    // [in] int           global line number
    //
  void set_fields (PTR_CNTXT ctxt, statement_tag varnt, PTR_SYMB sym,
		   PTR_EXPR e1, PTR_EXPR e2, PTR_EXPR e3, int gl, int ll) {
    if (ctxt)
      context = ctxt ;
    
    stmt_variant (varnt) ;
    g_line = gl ;
    l_line = ll ;
    control_parent = 0 ;
    label.ftn = 0 ;
    idx = 0 ;
    trans = 0 ;
    node.fields.symbol = sym ;
    node.fields.e1 = e1 ;
    node.fields.e2 = e2 ;
    node.fields.e3 = e3 ;
    node.fields.cmnt_ptr = 0 ;
    node.fields.lbl_ptr = 0 ;
    node.fields.bl_ptr1 = 0 ;
    node.fields.bl_ptr2 = 0 ;
  }

    ////////////////////////////////////
    //
    // This function makes this statement a non-recursive copy of
    // the (leaf) Statement passed as reference into this
    // routine. It remains to be determined how control parents
    // (i.e. non-leaf statements) should be copied. For example, a DO 100
    // statement should be modified to DO <new_label> so that label
    // conflicts don't arise.
    //
    // The control parent of this Statement is assumed to be the same
    // as the control parent of the reference Statement, and must be
    // assigned otherwise from the calling routine if necessary.
    //
    // This Statement is given the same line number as the reference
    // Statement, the same variant field, and the same label. The
    // Expressions of this Statement are full copies of the Expressions
    // in the reference Statement.
    //
    // [in] Statement&    reference statement
    //
  void set_fields (Statement& stat) ;

    ////////////////////////////////////
    //
    // Destructor.
    //
  ~Statement() ;

    ////////////////////////////////////
    //
    // Unparses this Fortran statement into a SourceListing object
    // which is passed into this routine by reference.
    //
    // [in] dialect        the dialect of Fortran being unparsed
    // [in] SourceListing& reference to the (created) SourceListing
    //
  void FortranUnparse (dialect, SourceListing&) ;

    ////////////////////////////////////
    //
    // Unparses this ANSI C statement into a SourceListing object
    // which is passed into this routine by reference.
    //
    // [in] SourceListing& reference to the (created) SourceListing
    //
  void CcUnparse (SourceListing&) ;

  void CcUnparseStatement (SourceListing&) ;

    ////////////////////////////////////
    //
    // Obligatory integer indexing function
    //
  int index()               { return id ; }
  // variant() provided by Tag class

    ////////////////////////////////////
    //
    // [out] int returns the global line number
    //
  int global_line_no ()     { return g_line ; }

    ////////////////////////////////////
    //
    // [out] int returns the local line number
    //
  int local_line_no ()      { return l_line ; }

    ////////////////////////////////////
    //
    // [out] Symbol* returns the associated Symbol
    //
  PTR_SYMB symbol ()        { return node.fields.symbol ; }

    ////////////////////////////////////
    //
    // [out] Statement* returns the Control Parent of this Statement
    //
    //
  PTR_STMT cntl_parent()    { return control_parent ; }

    ////////////////////////////////////
    //
    // Returns a pointer to the conditional expression in
    // an IF statement, etc.
    //
    // [out] ExpressionN* the conditional expression
    //
  PTR_EXPR condition ()     { return node.fields.e1 ; }

    ////////////////////////////////////
    //
    // Returns the list of PROC_HEDR, FUNC_HEDR, BLOCK_DATA
    // PROG_HEDR Statements which are defined by this
    // GLOBAL Statement. Note, there is one GLOBAL Statement
    // per File object.
    //
    // [out] PTR_BLOCK  the statments of the sub-program
    //
  PTR_BLOCK headers ()    { return node.fields.bl_ptr1 ; }

    ////////////////////////////////////
    //
    // Returns a list of Statements representing the body
    // of a loop if this statement is a DO loop, or some
    // similar construct which has a single set of control 
    // children.
    //
    // [out] PTR_BLOCK  the loop body
    //
  PTR_BLOCK loop_body ()    { return node.fields.bl_ptr1 ; }

    ////////////////////////////////////
    //
    // Returns the statements for which this Statement is
    // the control parent. Use this to access the StmtList
    // representing the body of a SUBROUTINE or FUNCTION
    //
    // [out] PTR_BLOCK  the statments of the sub-program
    //
  PTR_BLOCK statements ()    { return node.fields.bl_ptr1 ; }

    ////////////////////////////////////
    //
    // Returns the THEN part of an IF () THEN statement
    //
    // [out] PTR_BLOCK  the THEN block
    //
  PTR_BLOCK then_part ()    { return node.fields.bl_ptr1 ; }

    ////////////////////////////////////
    //
    // Returns the ELSE part of an IF () THEN ... ELSE statement
    //
    // [out] PTR_BLOCK  the ELSE block
    //
  PTR_BLOCK else_part ()    { return node.fields.bl_ptr2 ; }

    ////////////////////////////////////
    //
    // Returns the ScopeNode representing all Symbols declared
    // by this Statement.
    //
    // [out] ScopeNode*  the scope (may be NULL).
    //
  PTR_SCOPE scoping ()      { return scope ; }

    ////////////////////////////////////
    //
    // Simply returns Expression 1 from the Statement structure
    //
    // [out] Expression* the expression
    //
  PTR_EXPR expr1 ()         { return node.fields.e1 ; }

    ////////////////////////////////////
    //
    // Simply returns Expression 2 from the Statement structure
    //
    // [out] Expression* the expression
    //
  PTR_EXPR expr2 ()         { return node.fields.e2 ; }

    ////////////////////////////////////
    //
    // Simply returns Expression 2 from the Statement structure
    //
    // [out] Expression* the expression
    //
  PTR_EXPR expr3 ()         { return node.fields.e3 ; }

    ////////////////////////////////////
    //
    // Returns a pointer to universe of ftn labels defined
    // for the sub-program rooted at this Statement.
    //
  Universe<FtnLabel>* ftnLabels() { return node.procedure.labels ; }

    ////////////////////////////////////
    //
    // Returns a pointer to universe of C/C++ labels defined
    // for the sub-program rooted at this Statement.
    //
  LabelTable* ccLabels() { return node.function.labels ; }

    ////////////////////////////////////
    //
    // Returns an integer which is the label of goto if this
    // Statement is a logical IF statement.
    //
  int gotoLabel () ;

    ////////////////////////////////////
    //
    // Annotates this Statement with a pointer to a transformation
    // record describing the transformation applied to this node.
    //
  void note_transform (PTR_TRANS t) { trans = t ; }

    ////////////////////////////////////
    //
    // Returns a pointer to a TransHist object describing
    // describing the transformation applied to this node.
    // May return a null pointer if there was no transformation.
    //
  PTR_TRANS transformation ()       { return trans ; }

    ////////////////////////////////////
    //
    // Returns a non-zero value if the argument Statement S
    // is a direct descendent of 'this' statement, i.e. if
    // 'this' is an ancestor of S. Returns zero if this
    // relationship does not hold, and also if S == this,
    // or if either are null pointers.
    //
  int is_ancestor_of (PTR_STMT S) ;

    ////////////////////////////////////
    //
    // Returns a non-zero value if the loop nest rooted
    // at the outer-level by this statement, and at the
    // inner level by the argument S, is perfectly nested.
    // Perfect nesting occurs when all DO statements
    // except the inner-most contain only a single DO statement.
    // The inner-most loop can contain any statements, including
    // further loops.
    //
  int is_perfectly_nested_upto (PTR_STMT S) ;

    ////////////////////////////////////
    //
    // Returns a non-zero value if the loop nest rooted
    // at the outer-level by this statement is perfectly nested.
    // Perfect nesting occurs when all DO statements
    // except the inner-most contain only a single DO statement.
    // The inner-most loop can contain any statements, including
    // further loops.
    //
  int is_perfectly_nested () ;

    ////////////////////////////////////
    //
    // Returns a count of the number of statements in the
    // body of this statement. This is NOT a recursive count,
    // and only includes statements at the first level down.
    // Hence, a DO statement within the body only counts as one
    // statements regardless of the statements within that DO.
    // ENDDO and ENDWHILE statements are not counted.
    //
  int body_size () ;

    ////////////////////////////////////
    //
  PTR_STMT nextInnerLoop () ;

    ////////////////////////////////////
    //
  PTR_STMT nextOuterLoop () ;

    ////////////////////////////////////
    //
    // Transforms labelled DO statements into DO...ENDDO
    // constructs.
    //
    // All DO statements contained within this statement
    // are converted using a recursive search of the statement
    // tree structure.
    //
  void enforce_enddo () ;

    ////////////////////////////////////
    //
    // Returns a non-zero value if this statement is a DO loop
    // and it terminates with an `enddo'.
    //
  int is_do_end_loop () ;

    ////////////////////////////////////
    //
  int CcIndentation () ;
};

/*--------------------------------------------------------------

CLASS
    StmtList

    Linear list of Statements

DESCRIPTION

    This class implements a linear list of Statements, effectively
    representing a block in a block-structured language. The
    Statements are doubly-linked (using a dlist template) to
    permit bi-directional iteration.

----------------------------------------------------------------*/

class StmtList : public dlist<int,Statement> {

friend class StmtListIterator ;

public:

    ////////////////////////////////////
    //
    // Constructor.
    //
  StmtList() : dlist<int,Statement> ()  { } 

    ////////////////////////////////////
    //
    // Conversion constructor.
    //
  StmtList(dlist<int,Statement>& sl) : dlist<int,Statement> (sl)  { } 

    ////////////////////////////////////
    //
    // Copy function copies this statement list recursively
    // using S as the initial control parent of the list.
    //
  StmtList& copy (PTR_STMT S) ;

    ////////////////////////////////////
    //
    // Set the control parent of all Statements in this
    // StmtList to the given Statement argument
    //
  void set_block_parent (PTR_STMT) ;

  // GROUP: unparsing functions for C and C++ statement lists

    ////////////////////////////////////
    //
  void CcUnparseStmtList (SourceListing&) ;

    ////////////////////////////////////
    //
  void CcUnparseCompoundStmt (SourceListing&, int, PTR_STMT) ;

};

/*--------------------------------------------------------------

CLASS
    StmtListIterator

    Iterator class for lists of Statements

DESCRIPTION

    Implements a class specifically for iterating conveniently
    through a list of Statements.

    As with all our iterator class, one passes a member of
    the list class into the constructor of the iterator to
    initialise the iteration process.

    The () operator is used to step through the Statement list.
----------------------------------------------------------------*/

class StmtListIterator : public dlist_iterator<int,Statement> {

public:

    ////////////////////////////////////
    //
    // Constructor.
    //
  StmtListIterator (StmtList& sl) : dlist_iterator<int,Statement> (sl) { }

    ////////////////////////////////////
    //
    // Constructor.
    //
  StmtListIterator (StmtList* sl) : dlist_iterator<int,Statement> (sl) { }

    ////////////////////////////////////
    //
    // Positions the iterator at the transition between
    // declaration statements and executable statements.
    // After a call to this function, the iterator is
    // positioned so that the next Statement returned by
    // the head() function is the first executable, or
    // zero if there are no executables.
    //
    // A call to insert(S) following a call to this function
    // will insert statement S after the last declaration
    // and before the first executable.
    //
    // If there are no declaration statements in the statement list
    // then the iterator is positioned prior to the first
    // statement.
    //
    // If the statement list is empty the iterator is positioned
    // such that the () operator returns zero, but the insert(S)
    // operator will still correctly insert statement S into
    // the empty list.
    //
  void goto_first_exec () ;

    ////////////////////////////////////
    //
    // Positions the iterator at the latest point where
    // declaration statements are permitted.
    // After a call to this function, the iterator is
    // positioned so that the next Statement returned by
    // the head() function is either the first executable, or
    // a DATA, COMMON, EQUIVALENCE, or BLOCKDATA statement
    // (or zero if there are no such statements in the list).
    //
    // A call to insert(S) following a call to this function
    // will insert declaration statement S at a syntactically
    // correct position.
    //
    // If the statement list is empty the iterator is positioned
    // such that the () operator returns zero, but the insert(S)
    // operator will still correctly insert statement S into
    // the empty list.
    //
  void goto_declarations () ;

    ////////////////////////////////////
    //
    // Split a loop nest L into two sequential nests L1 and L2,
    // each iterating over the same space and computing
    // in a semantically-equivalent way.
    //
    // If 'this' statement iterator is pointing at a loop
    // header (i.e. DO statement) then the loop nest
    // which has that statement as the outer-most loop
    // declarator is split into two loop nests with
    // identical iteration spaces.
    //
    // The given loop nest is first normalised to transform
    // label-terminated DO statements into DO/ENDDO constructs.
    // This avoids ambiguity when loop nests are duplicated.
    //
    // The control parent of Statement S given as argument
    // identifies the inner-most DO statement of the loop
    // nest to be split. It has a loop body SL containing
    // the argument statement. Let L1 be the DO nest from
    // the current statement to the control parent of S.
    //
    // The first loop contains all statements from the first
    // statement of SL up to, but not including S. The second
    // loop contains all statements from, and including, S
    // up to the end of SL.
    //
    // If either of these two statement lists is empty, no
    // split is actually made. Otherwise, an additional and
    // identical loop nest L2 is constructed from L1
    // and inserted after L1.
    //
    //  [in]  Statement* pointer to split point
    //  [out] returns 1 for success and 0 for failure
    //
  int split_loop_nest (PTR_STMT S) ;

    ////////////////////////////////////
    //
    // Cuts out the statement sub-list from the mark to
    // the current position (or end of list).
    //
  StmtList* cut ()
  {
    // void* t = dlist_iterator<int, Statement>::cut() ;
    // return (StmtList*)t ;

    return (StmtList*)dlist_iterator<int, Statement>::cut() ;
  }

    ////////////////////////////////////
    //
    //
    //
    //
  void insert (PTR_STMT s, PTR_STMT cp)
  {
    dlist_iterator<int,Statement>::insert (s) ;
    s->control_parent = cp ;
  }

};

#endif // _statements_h_


