/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 | Copyright (C) University of Indianna & IRISA, Rennes, 1991.        |
 |  This source code is derived from the original db_unp.c file from  |
 |  the 1991 Sigma distribution.                                      |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// $Id: .cc,v 1.11 1999/08/16 08:02:51 npt Exp $

/* This file currently supports only Fortran unparsing */

#include "Octave/Base/Unparse.h"
#include "Octave/Base/ParseContext.h"
#include "Octave/Fortran/Fortran.h"
#include "Octave/Fortran/FortranLexer.h"

/* NOTES:
   
   1. These routines should make a distinction between unparsing
      declarators and types (from a user-interface perspective.

*/


#define MODULE_TESTING 1

// Function prototypes for simple Fortran unparsing

//  1. The principal syntax-directed unparsing routines

static void FortranUnparseUnit  (SourceListing&, PTR_STMT) ;
static void FortranUnparseBlock (SourceListing&, PTR_BLOCK) ;
static void FortranUnparseStatement (SourceListing&, PTR_STMT) ;
static void FortranUnparseLine (LineBuffer&, PTR_STMT) ;
static void FortranUnparseType (LineBuffer&, PTR_TYPE, int) ;
static void gen_loop_header (LineBuffer&, char *, PTR_STMT) ;
static void generate_branch (LineBuffer&, statement_tag, char *, PTR_STMT) ;

static int cop_prec (expression_tag) ;
static char* cop_name (expression_tag) ;

//  2. Support routines for textual alignment and output

static int computeIndentation (PTR_STMT) ;
static int computeIndentation (PTR_SCOPE) ;

/* Static variables to guide the unparsing process and to store the
 * intermediate strings.
 */

static int in_param = 0;	// set if unparsing the parameter statement
static int in_impli = 0;	// set if unparsing the implicit statement
static int in_sequence = 0 ;    // set if unparsing a dospec expression
static PTR_CMNT cmnt = NULL;	// point to chain of comment list
static int print_comments = 1;  // 0 if no comments
static int in_var_decl = 0;     // cntls printing of var decls
static int with_kind = 0;       // 0 if var decl has no kind param, 1 otherwise
static int no_var_leng_spec = 0;// cntls printing of var length spec in decls
static PTR_TYPE decl_type = 0 ; // keeps track of type node in var decl.

/*
 * Ascii names for Fortran operators
 */

static
char *fop_name[] = {
	" .eq. ",
	" .lt. ",
	" .gt. ",
	" .ne. ",
	" .le. ",
	" .ge. ",
	"+",
	"-",
	" .or. ",
	"*",
	"/",
	"",
	" .and. ",
	"**",
	"",
	"//",
	" .xor. ",
	" .eqv. ",
	" .neqv. "
};


/*
 * Precedence table of operators for Fortran
 */
static char f_precedence[] = {
	5,		/* .eq. */
	5,		/* .lt. */
	5,		/* .gt. */
	5,		/* .ne. */
	5,		/* .le. */
	5,		/* .ge. */
	3,		/*  +   */
	3,		/*  -   */
	8,		/* .or. */
	2,		/*  *   */
	2,		/*  /   */
	0,		/* none */
	7,		/* .and. */
	1,		/*  **  */
	0,		/* none */
	4,		/*  //  */
	8,		/* .xor. */
	9,		/* .eqv. */
	9		/* .neqv. */
};

/*
 * Fortran type names in ascii form
 */
static
char *ftype_name[] = {
	"integer",
	"real",
	"double precision",
	"character",
	"logical",
	"character",
	"gate",
	"event",
	"sequence",
	"",
	"",
	"",
	"",
	"complex"
};


/*---------------------------------------------------------------------+
 |                                                                     |
 |                     SourceListing::Unparse                          |
 |                                                                     |
 +---------------------------------------------------------------------*/

void SourceListing::unparse (PTR_STMT stmt, dialect dlt) {
  if (!stmt)
    return ;

  lang = stmt->context->OctLanguage() ;

  switch (lang) {
  case F77:
    stmt->FortranUnparse (dlt, *this) ; 
    break ;
  case AnsiC:
  case CPlusPlus:
    stmt->CcUnparse (*this) ;
    break ;
  default:
    break ;
  }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                    Statement::FortranUnparse                        |
 |                                                                     |
 +---------------------------------------------------------------------*/

void Statement::FortranUnparse (dialect dlt, SourceListing& src)
{
  if (stmt_variant() == GLOBAL)
    FortranUnparseBlock (src, node.fields.bl_ptr1) ;
  else {
    FortranUnparseUnit (src, this) ;
  }
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                     LineBuffer::PrefixLine                          |
 |                                                                     |
 +---------------------------------------------------------------------*/

void LineBuffer::PrefixLine (int lb, int tb, int isCmnt, int isDir, 
			     PTR_STMT s) {
  isComment = isCmnt ;
  isDirective = isDir ;
  label = lb ;
  tabs = tb ;
  stmt = s ;
  pos = 0 ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                    LineBuffer::CcStartLine                          |
 |                                                                     |
 +---------------------------------------------------------------------*/

void LineBuffer::CcStartLine (int tb, PTR_STMT s, PTR_CC_LABEL lab)
{
  isComment = 0 ;
  isDirective = 0 ;
  label = 0 ;
  tabs = tb ;
  stmt = s ;
  pos = 0 ;
  if (lab)
    {
      TabIndent (tb-1) ;
      (*this) << lab->index() ;
      (*this) << ": " ;
    }
  else
    TabIndent (tb) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                      static void emitLine                           |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void emitLine (SourceListing& src, int label, int tabs, char *text, 
		      PTR_STMT stmt) {
  LineBuffer buff ;
  
  Comment *cptr = stmt->node.fields.cmnt_ptr ;

  while (cptr && cptr->type == FULL) {
    buff.PrefixLine (0, 0, 1, 0, stmt) ;
    buff << 'C' ;
    buff.accept_raw (cptr->cstring()) ;
    buff << '\n' ;
    src.appendBuffer (buff) ;
    cptr = cptr->next ;
  }

  buff.PrefixLine (label, tabs, false, false, stmt) ;
  buff << text ;

  if (cptr && cptr->type != FULL) {
    buff << " ! " ;
    buff << cmnt->cstring() ;
  }
  buff << '\n' ;

  src.appendBuffer (buff) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |               static void FortranUnparseBlock                       |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void FortranUnparseBlock (SourceListing& src, PTR_BLOCK block) 
{
  if (!block)
    return ;

  StmtListIterator next_unit (block) ;
  PTR_STMT unit ;

  while ( unit = next_unit() )
    FortranUnparseUnit (src, unit) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |               static void FortranUnparseUnit                        |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void FortranUnparseUnit (SourceListing& src, PTR_STMT block) 
{ 
  FortranUnparseStatement (src, block) ;
  int tabs = computeIndentation (block) ;
  
  if (   block->stmt_variant () != CDOALL_NODE 
      && block->stmt_variant () != SDOALL_NODE) 
  {
    StmtListIterator next_stmt (block->node.fields.bl_ptr1) ;
    PTR_STMT stmt ;

    while (stmt = next_stmt () )
    {
      if (stmt->stmt_variant () != CONTROL_END)
	FortranUnparseUnit (src, stmt) ;
      else {
	switch (block->stmt_variant ()) {
	case FOR_NODE:
	case PARDO_NODE:
	case PDO_NODE:
	case WHILE_NODE:
	  if (!block->node.fields.lbl_ptr) {
	    if (block->stmt_variant () == PARDO_NODE)
	      emitLine (src, 0, tabs, "end parallel do", stmt) ;
	    else if (block->stmt_variant () == PDO_NODE)
	      emitLine (src, 0, tabs, "end pdo", stmt) ;
	    else
	      emitLine (src, 0, tabs, "end do", stmt) ;
	  }
	  break;

	case IF_NODE: 
	case ELSEIF_NODE:
	  if (block->node.fields.bl_ptr2)
	  {
	    if ((block->node.fields.bl_ptr2->hd ()->stmt_variant ())!=ELSEIF_NODE)
	      emitLine (src, 0, tabs, "else", stmt);
	  }
	  else
	    emitLine (src, 0, tabs, "end if", stmt) ;
	  break;

	case WHERE_BLOCK_STMT:
	  if (block->node.fields.bl_ptr2)
	    emitLine (src, 0, tabs, "elsewhere", stmt) ;
	  else
	    emitLine (src, 0, tabs, "end where", stmt) ;
	  break;

	case CASE_NODE:
	{
	  char bf [50] ;
	  if (block->node.fields.symbol)
	    sprintf (bf, "end select %s", (block->node.fields.symbol->ident ()) );
	  else
	    sprintf (bf, "end select " );

	  emitLine (src, 0, tabs, bf, stmt) ;
	  break;
	}
	case SWITCH_NODE:
	  break;

	case PROG_HEDR:
	case PROC_HEDR:
	case FUNC_HEDR:
	case BLOCK_DATA:
	  emitLine (src, 0, tabs, "end", stmt) ;
	  break;

	case MODULE_STMT:
	{
	  char bf [50] ;
	  sprintf (bf, "end module %s", block->node.fields.symbol->ident()) ;
	  emitLine (src, 0, tabs, bf, stmt) ;
	  break;
	}
	case INTERFACE_STMT:
	  emitLine (src, 0, tabs, "end interface", stmt) ;
	  break;

	case STRUCT_DECL:
	{
	  char bf [50] ;
	  sprintf (bf, "end type %s", block->node.fields.symbol->ident ()) ;
	  emitLine (src, 0, tabs, bf, stmt) ;
	  break;
	}
	case PARREGION_NODE:
	  emitLine (src, 0, tabs, "end parallel", stmt) ;
	  break;

	case PARSECTIONS_NODE:
	  emitLine (src, 0, tabs, "end parallel sections", stmt) ;
	  break;

	case PSECTIONS_NODE:
	  emitLine (src, 0, tabs, "end psections", stmt) ;
	  break;

	case SINGLEPROCESS_NODE:
	  emitLine (src, 0, tabs, "end single process", stmt) ;
	  break;

	case CRITSECTION_NODE:
	  emitLine (src, 0, tabs, "end critical section", stmt) ;
	  break;

	default:
	  break; 
	}
      }
    }
    
    StmtListIterator next_stmt2 (block->node.fields.bl_ptr2) ;

    while ( stmt = next_stmt2 () )
    {
      if (stmt)
      {
	if (stmt->stmt_variant () != CONTROL_END)
	  FortranUnparseUnit (src, stmt) ;
	else 
	{
	  if (block->stmt_variant () == PDO_NODE) 
	    emitLine (src, 0, tabs, "end extended", stmt) ;
	  if (block->stmt_variant () == PSECTIONS_NODE)
	    emitLine (src, 0, tabs, "end extended", stmt) ;
	  if (block->stmt_variant () == WHERE_BLOCK_STMT)
	    emitLine (src, 0, tabs, "end where", stmt) ;
	  if ((block->stmt_variant () == IF_NODE) || (block->stmt_variant () == ELSEIF_NODE))
	    emitLine (src, 0, tabs, "end if", stmt) ;
	}
      }
    }
  }
  else { /* deal with CDOALL nodes etc. etc. */
    StmtListIterator next_stmt (block->node.fields.bl_ptr2) ;
    PTR_STMT stmt ;

    while (stmt = next_stmt () )
    {
      if (stmt->stmt_variant () != CONTROL_END)
	FortranUnparseUnit (src, stmt) ;
      else {
	if (!block->node.fields.lbl_ptr)
	  emitLine (src, 0, tabs, "loop", stmt) ;
      }
    }
     
    StmtListIterator next_stmt1 = block->node.fields.bl_ptr1 ;
    
    while (stmt = next_stmt1 () )
    {
      if (stmt->stmt_variant () != CONTROL_END)
	FortranUnparseUnit (src, stmt) ;
      else {
	if (block->stmt_variant () == CDOALL_NODE)
	  emitLine (src, 0, tabs, "end cdoall", stmt) ;
	else
	  emitLine (src, 0, tabs, "end sdoall", stmt) ;
      }
    }
    
  }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                 static void FortranUnparseType                      |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void FortranUnparseType (LineBuffer& buff, PTR_TYPE ptype, int def)
{
  if (!ptype)
    return ;
  
  if (def == 0 && ptype->name) { /* print the type name */
    buff << ptype->name->ident() ;
    return ;
  }   
  
  switch (ptype->type_variant ()) {
  case T_INT   :
  case T_FLOAT :
  case T_DOUBLE:
  case T_CHAR  :
  case T_BOOL  :
  case T_STRING:
  case T_COMPLEX:
    buff << ftype_name[ptype->type_variant () - T_INT] ;
    break;
  case T_GATE:
    buff << "gate" ;
    break;
  case T_EVENT:
    buff << "event" ;
    break;
  case T_SEQUENCE:
    buff << "sequence" ;
    break;
  case T_ARRAY :
    FortranUnparseType (buff, ptype->node.ar_decl.base_type, 0);
    break;
  case T_DERIVED_TYPE:
    buff << "type (" ;
    buff << ptype->name->ident () ;
    buff << ")" ;
    break;
  case T_POINTER:
    FortranUnparseType (buff, ptype->node.fields.base_type, 0);
    break;
    
    default :
      return ;
  }
  return ;
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                 static void gen_loop_header                         |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void gen_loop_header (LineBuffer& buff, char *looptype, PTR_STMT pbf)
{
  char	label[7];
  
  buff << looptype ;
  if ((pbf->stmt_variant () == PARDO_NODE) || (pbf->stmt_variant () == PDO_NODE))
    if (pbf->node.for_node.where_cond)
    {
      buff << " ( " ;
      FortranUnparseExpression (buff, pbf->node.for_node.where_cond);
      buff << " ) " ;
    }
  if (pbf->node.for_node.doend) {
    sprintf (label,"%d ",pbf->node.for_node.doend->index ());
    buff << label ;
  }
  buff << pbf->node.for_node.control_var->ident () ;
  buff << " = " ;
  FortranUnparseExpression (buff, pbf->node.for_node.range->node.binary_op.l_operand);
  buff << ", " ;
  FortranUnparseExpression (buff, pbf->node.for_node.range->node.binary_op.r_operand);
  if (pbf->node.for_node.increment) {
    buff << " , " ;
    FortranUnparseExpression (buff, pbf->node.for_node.increment);
  }
}



/*---------------------------------------------------------------------+
 |                                                                     |
 |                   static void generate_branch                       |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void generate_branch (LineBuffer& buff, statement_tag branch_tag, 
			     char *branch_type, PTR_STMT pbf)
{
  buff << branch_type ;
  buff << '(';
  FortranUnparseExpression (buff, pbf->node.if_node.condition) ;
  buff << ')';
  if (branch_tag != WHERE_BLOCK_STMT)
    buff << " then" ;
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |               static void FortranUnparseExpression                  |
 |                                                                     |
 +---------------------------------------------------------------------*/

void FortranUnparseExpression (LineBuffer& buff, PTR_EXPR pllnd)
{
  if (pllnd == 0)
    return ;

  switch (pllnd->expr_variant ()) {
  case INT_VAL	:
  { char sb[64];
    
    sprintf (sb, "%d", pllnd->node.ival);
    buff << sb ;
    break;
  }

  case LABEL_REF:
  { char sb[64];
    
    sprintf (sb, "%d", pllnd->node.label_list.lab_ptr->index ());
    buff << sb ;
    break;
  }

  case FLOAT_VAL	:
  case DOUBLE_VAL	:
  case STMT_STR	:
    buff << (*pllnd->node.string_val)() ;
    break;
		  
  case STRING_VAL	:
    {
      // Must scan string for hollerith delimiters and convert back to holleriths
      // This is because old-style (i.e. Fortran 66) holleriths within FORMAT
      // or DATA statements are delimited with HOLL_DELIM characters during
      // lexical analysis to enable optional conversion to CHARACTER strings
      // (i.e. Fortran 77).
      // At present this is not done for FORMAT and DATA statements, so we must
      // convert such holleriths back to Fortran 66 hollerith syntax, taking
      // care to implode pairs of quotation characters back to single quotation
      // characters.

      char *str = (*pllnd->node.string_val)() ;
      while (str && *str) {
	if (*str == HOLL_DELIM) {
	  int hcnt, lastNotQuote ;
	  char hbuf [256] ;
	  char obuf [256] ;

	  for (hcnt = 0, lastNotQuote = 1, ++str ; *str && *str != HOLL_DELIM ; str++) {
	    if ( lastNotQuote || *str != '\'' )
	      hbuf [hcnt++] = *str ;
	    lastNotQuote = *str != '\'' ;
	  }
	  if (*str == HOLL_DELIM) ++str ;
	  hbuf [hcnt] = '\0' ;
	  sprintf (obuf, "%dH%s", hcnt, hbuf) ;
	  buff << obuf ;
	}
	else
	  buff << *str++ ;
      }
      break;
    }

  case COMPLEX_VAL	:
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << ',';
    FortranUnparseExpression (buff, pllnd->node.fields.e2);
    buff << ')';
    break;

  case OCTAL_VAL:
    buff << (*pllnd->node.string_val)() ;
    break ;

  case HEX_VAL:
    buff << (*pllnd->node.string_val)() ;
    break ;

  case KEYWORD_VAL	:
    buff << (*pllnd->node.string_val)() ;
    break;

  case KEYWORD_ARG	:
    buff << (char *)(pllnd->node.fields.e1) ;
    buff << "=" ;
    buff << (char *)(pllnd->node.fields.e2) ;
    break;

  case BOOL_VAL	:
    if (pllnd->node.bval)
      buff << ".true." ;
    else 
      buff << ".false." ;
    break;

  case CHAR_VAL	:
    if (! in_impli)
      buff << '\'';
    buff << pllnd->node.cval;
    if (! in_impli)
      buff << '\'';
    break;

  case CONST_REF:
  case ENUM_REF:
  case TYPE_REF:
  case INTERFACE_REF:
    buff << pllnd->node.fields.symbol->ident () ;
  /*--- this does not seem to have any purpose ---
    if ((in_var_decl) && (pllnd->node.fields.symbol->type))
      if ((pllnd->node.fields.symbol->type->node.fields.ranges != 0) &&
	  !no_var_leng_spec && !in_param)
	FortranUnparseExpression (buff, pllnd->node.fields.symbol->type->node.fields.ranges) ;
   -----------*/
    break;

  case VAR_REF:
  {
    buff << pllnd->node.fields.symbol->ident () ;
    
    /* Unparse the length spec only if we are in a variable declaration and
     * this var has type different from that in the declaring TYPE_OP, and
     * the type in the TYPE_OP does not have a kind parameter.
     */

    if (in_var_decl && !no_var_leng_spec && !with_kind && !in_param)
      {
	PTR_TYPE t = pllnd->node.fields.symbol->type ;
	PTR_EXPR len = 0 ;

	if (t->type_variant() == T_ARRAY)
	  t = t->node.ar_decl.base_type ;

	if (t && t != decl_type)
	  {
	    len = t->node.fields.ranges ;
	    
	    if (len)
	      FortranUnparseExpression (buff, len) ;
	  }
      }
    break;
  }

  case ARRAY_REF:
  {
    buff << pllnd->node.array_ref.symbol->ident () ;
    if (pllnd->node.array_ref.index)
      {
	int nvls = no_var_leng_spec  ;
	
	buff << '(';
	no_var_leng_spec = 1 ;
	if (pllnd->node.array_ref.index->node.fields.e1 == pllnd)
	  buff << "$cyclic-expr$" ;
	else
	  FortranUnparseExpression (buff, pllnd->node.array_ref.index) ;
	no_var_leng_spec = nvls ;
	buff << ')' ;
      }

    if (pllnd->node.array_ref.array_elt)
      {
	int nvls = no_var_leng_spec  ;
	
	buff << '(';
	no_var_leng_spec = 1 ;
	if (pllnd->node.array_ref.index->node.fields.e2 == pllnd)
	  buff << "$cyclic-expr$" ;
	else
	  FortranUnparseExpression (buff, pllnd->node.array_ref.array_elt) ;
	no_var_leng_spec = nvls ;
	buff << ')' ;
      }
    
    if (in_var_decl && !no_var_leng_spec && !with_kind && !in_param)
      {
	PTR_TYPE t = pllnd->node.fields.symbol->type ;
	
	if (t)
	  {
	    PTR_TYPE bt = t->node.fields.base_type ;
	    
	    if (bt && bt != decl_type && bt->node.fields.ranges)
	      FortranUnparseExpression (buff, bt->node.fields.ranges) ;
	  }
      }
    break;
  }

  case ARRAY_OP	:
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.fields.e2);
    buff << ')';
    break;

  case RECORD_REF	:
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << "%" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e2);
    break;

  case STRUCTURE_CONSTRUCTOR	:
    buff << pllnd->node.fields.symbol->ident () ;
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << ')';
    break;

  case CONSTRUCTOR_REF	:
    buff << "(/" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << "/)" ;
    break;

  case ACCESS_REF	:
    FortranUnparseExpression (buff, pllnd->node.access_ref.access);
    if (pllnd->node.access_ref.index != NULL) {
      buff << '(';
      FortranUnparseExpression (buff, pllnd->node.access_ref.index);
      buff << ')';
    }
    break;

  case OVERLOADED_CALL:
    break;

  case CONS	:
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << ", " ;
    FortranUnparseExpression (buff, pllnd->node.fields.e2);
    break;

  case ACCESS	:
    FortranUnparseExpression (buff, pllnd->node.access.array);
    buff << ", FORALL=(" ;
    buff << pllnd->node.access.control_var->ident () ;
    buff << '=';
    FortranUnparseExpression (buff, pllnd->node.access.range);
    buff << ')';
    break;

  case IOACCESS	:
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.ioaccess.array);
    buff << ", " ;
    buff << pllnd->node.ioaccess.control_var->ident () ;
    buff << '=';
    FortranUnparseExpression (buff, pllnd->node.ioaccess.range);
    buff << ')';
    break;

  case PROC_CALL	:
  case FUNC_CALL	:
    buff << pllnd->node.proc.symbol->ident () ;
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.proc.param_list);
    buff << ')';
    break;

  case EXPR_LIST	:
    FortranUnparseExpression (buff, pllnd->node.list.item);
    if (in_param) {
      buff << "=" ;
      FortranUnparseExpression (buff, pllnd->node.list.item->node.const_ref.symbol->
				node.const_value) ;
    }
    if (pllnd->node.list.next) {
      buff << ", " ;
      FortranUnparseExpression (buff, pllnd->node.list.next);
    }
    break;

  case EQUI_LIST	:
    buff << '(';
    FortranUnparseExpression (buff, pllnd->node.list.item);
    buff << ')';
    if (pllnd->node.list.next) {
      buff << ", " ;
      FortranUnparseExpression (buff, pllnd->node.list.next);
    }
    break;

  case COMM_LIST	  :
  case NAMELIST_LIST:
    if (pllnd->node.fields.symbol) {
      buff << '/';
      buff << pllnd->node.fields.symbol->ident () ;
      buff << "/ ";
    }
    FortranUnparseExpression (buff, pllnd->node.list.item);
    if (pllnd->node.list.next) {
      buff << ", " ;
      FortranUnparseExpression (buff, pllnd->node.list.next);
    }
    break;

  case VAR_LIST	 :
  case RANGE_LIST	 :
  case CONTROL_LIST:
    FortranUnparseExpression (buff, pllnd->node.list.item);
    if (pllnd->node.list.next) {
      buff << ", " ;
      FortranUnparseExpression (buff, pllnd->node.list.next);
    }
    break;

  case DDOT	:
    if (pllnd->node.binary_op.l_operand)
      FortranUnparseExpression (buff, pllnd->node.binary_op.l_operand);

    if (in_impli)
      buff << '-' ;
    else if (in_sequence)
      buff << ',' ;
    else
      buff << ':' ;

    if (pllnd->node.binary_op.r_operand) 
      FortranUnparseExpression (buff, pllnd->node.binary_op.r_operand);
    break;

  case DEFAULT_EXPR:
    buff << "default" ;
    break;

  case DEF_CHOICE	:
  case SEQ	:
    in_sequence = 1 ;
    FortranUnparseExpression (buff, pllnd->node.seq.ddot);
    in_sequence = 0 ;
    if (pllnd->node.seq.stride) {
      buff << ',';
      FortranUnparseExpression (buff, pllnd->node.seq.stride);
    }
    break;

  case SPEC_PAIR	:
    FortranUnparseExpression (buff, pllnd->node.spec_pair.sp_label);
    buff << '=';
    FortranUnparseExpression (buff, pllnd->node.spec_pair.sp_value);
    if (in_var_decl)
      with_kind = 1 ;
    break;
    
  case EQV_OP:
  case NEQV_OP:
  case EQ_OP:
  case LT_OP:
  case GT_OP:
  case NOTEQL_OP:
  case LTEQL_OP:
  case GTEQL_OP:
  case ADD_OP:
  case SUBT_OP:
  case OR_OP:
  case MULT_OP:
  case DIV_OP:
  case MOD_OP:
  case AND_OP:
  case EXP_OP:
  case CONCAT_OP:
    {
      int i = pllnd->expr_variant () - EQ_OP, j;
      PTR_EXPR p;
      int num_paren = 0;
      
      p = pllnd->node.binary_op.l_operand;
      j = p->expr_variant () ;
      if (IS_BIN_OP(j) && f_precedence[i] < f_precedence[j-EQ_OP]) {
	num_paren++;
	buff << '(';
      }
      FortranUnparseExpression (buff, p);
      if (num_paren) {
	buff << ')';
	num_paren--;
      }
      buff << fop_name[i] ; /* print the op name */
      p = pllnd->node.binary_op.r_operand;
      j = p->expr_variant ();
      if (IS_BIN_OP(j) && f_precedence[i] <= f_precedence[j-EQ_OP]) {
	num_paren++;
	buff << '(';
      }
      FortranUnparseExpression (buff, p);
      if (num_paren) {
	buff << ')';
	num_paren--;
      }
      break;
    }
  
  case MINUS_OP	:
/* Was before */
/*    
    buff << " -(" ;
    FortranUnparseExpression (buff, pllnd->node.unary_op.operand);
    buff << ')'; 
*/
    /* fgg changed on 2000/11/30 */
    buff << "-" ;
    FortranUnparseExpression (buff, pllnd->node.unary_op.operand);
    break;
    
  case UNARY_ADD_OP	:
    buff << " +(" ;
    FortranUnparseExpression (buff, pllnd->node.unary_op.operand);
    buff << ')';
    break;

  case NOT_OP	:
    buff << " .not. (" ;
    FortranUnparseExpression (buff, pllnd->node.unary_op.operand);
    buff << ')';
    break;

  case PAREN_OP:
    buff << "(" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << ")" ;
  break ;

  case ASSGN_OP:
    buff << "=" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
  break ;

  case STAR_RANGE	:
    buff << " * " ;
    break;

  case IMPL_TYPE:
    FortranUnparseType (buff, pllnd->type, 1) ;

    if (pllnd->type->node.fields.ranges)
      FortranUnparseExpression (buff, pllnd->type->node.fields.ranges) ;

    if (pllnd->node.fields.e1 != 0)
    {
      buff << " (" ;
      FortranUnparseExpression (buff, pllnd->node.fields.e1);
      buff << ")" ;
    }
    break;

  case ORDERED_OP :
    buff << "ordered " ;
    break;

  case EXTEND_OP :
    buff << "extended " ;
    break;

  case MAXPARALLEL_OP:
    buff << "max parallel = " ;
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    break;

  case PARAMETER_OP :
    buff << "parameter " ;
    break;

  case PUBLIC_OP :
    buff << "public " ;
    break;

  case PRIVATE_OP :
    buff << "private " ;
    break;

  case ALLOCATABLE_OP :
    buff << "allocatable " ;
    break;

  case DIMENSION_OP :
    buff << "dimension (" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << ")" ;
    break;

  case EXTERNAL_OP :
    buff << "external " ;
    break;

  case OPTIONAL_OP :
    buff << "optional " ;
    break;

  case IN_OP :
    buff << "intent (in) " ;
    break;

  case OUT_OP :
    buff << "intent (out) " ;
    break;

  case INOUT_OP :
    buff << "intent (inout) " ;
    break;

  case INTRINSIC_OP :
    buff << "intrinsic " ;
    break;

  case POINTER_OP :
    buff << "pointer " ;
    break;

  case SAVE_OP :
    buff << "save " ;
    break;
  case TARGET_OP :
    buff << "target " ;
    break;

  case LEN_OP :
    {
      buff << "*" ;
      PTR_EXPR length = pllnd->node.fields.e1 ;
      expression_tag etag = length ? length->expr_variant() : INT_VAL ;

      // Enclose the length-spec in parentheses is not an integer constant
      // and the length-spec is not already so enclosed.

      if (etag == PAREN_OP || etag == INT_VAL)
	FortranUnparseExpression (buff, length) ;
      else {
	buff << "(" ;
	FortranUnparseExpression (buff, length) ;
	buff << ")" ;
      }
      break;
    }

  case TYPE_OP :
    FortranUnparseType (buff, pllnd->type, 1);
    decl_type = pllnd->type ;
    if (pllnd->type->node.fields.ranges) {
      FortranUnparseExpression (buff, pllnd->type->node.fields.ranges);
    }
    break;

  case ONLY_NODE :
    buff << "only: " ;
    if (pllnd->node.fields.e1)
      FortranUnparseExpression (buff, pllnd->node.fields.e1);
    break;

  case DEREF_OP :
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    break;

  case RENAME_NODE :
    FortranUnparseExpression (buff, pllnd->node.fields.e1);
    buff << "=>" ;
    FortranUnparseExpression (buff, pllnd->node.fields.e2);
    break;

  case VARIABLE_NAME :
    buff << pllnd->node.fields.symbol->ident () ;
    break;

  // Following cases added for Convex Fortran Directives

  case ASSIGN_LOCK_DIR:
     buff << "assign_lock (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case FREE_LOCK_DIR:
     buff << "free_lock (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case BEGIN_ORDER_DIR:
     buff << "begin_order (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case END_ORDER_DIR:
     buff << "end order" ;
     break ;
  case BEGIN_TASKS_DIR:
     buff << "begin_tasks" ;
     break ;
  case END_TASKS_DIR:
     buff << "end_tasks" ;
     break ;
  case BEGIN_SECTION_DIR:
     buff << "begin_section (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case END_SECTION_DIR:
     buff << "end_section" ;
     break ;
  case FORCE_PARALLEL_EXT_DIR:
     buff << "force_parallel_ext" ;
     break ;
  case FORCE_PARALLEL_DIR:
     buff << "force_parallel" ;
     break ;
  case FORCE_VECTOR_DIR:
     buff << "force_vector" ;
     break ;
  case MAX_TRIPS_DIR:
     buff << "max_trips (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case NEXT_TASK_DIR:
     buff << "next_task" ;
     break ;
  case NO_PARALLEL_DIR:
     buff << "no_parallel" ;
     break ;
  case NO_PEEL_DIR:
     buff << "no_peel" ;
     break ;
  case NO_PROMOTE_TEST_DIR:
     buff << "no_promote_test" ;
     break ;
  case NO_RECURRENCE_DIR:
     buff << "no_recurrence" ;
     break ;
  case NO_SIDE_EFFECTS_DIR:
     buff << "no_side_effects (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case NO_VECTOR_DIR:
     buff << "no_vector" ;
     break ;
  case PEEL_ALL_DIR:
     buff << "peel_all" ;
     break ;
  case PEEL_DIR:
     buff << "peel" ;
     break ;
  case PROMOTE_TEST_ALL_DIR:
     buff << "promote_test_all" ;
     break ;
  case PROMOTE_TEST_DIR:
     buff << "promote_test" ;
     break ;
  case PREFER_PARALLEL_EXT_DIR:
     buff << "prefer_parallel_ext" ;
     break ;
  case PREFER_PARALLEL_DIR:
     buff << "prefer_parallel" ;
     break ;
  case PREFER_VECTOR_DIR:
     buff << "prefer_vector" ;
     break ;
  case PSTRIP_DIR:
     buff << "pstrip (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case ROW_WISE_DIR:
     buff << "row_wise (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case SCALAR_DIR:
     buff << "scalar" ;
     break ;
  case SELECT_DIR:
     buff << "select (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;
  case SYNCH_PARALLEL_DIR:
     buff << "synch_parallel" ;
     break ;
  case UNROLL_DIR:
     buff << "unroll" ;
     break ;
  case VSTRIP_DIR:
     buff << "vstrip (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;

  // Following cases added for CRI Fortran Directives

  case CRI_GUARD_DIR:
     buff << "GUARD" ;
     break ;

  case CRI_ENDGUARD_DIR:
     buff << "END GUARD" ;
     break ;

  case CRI_PARALLEL_DIR:
     buff << "PARALLEL" ;
     break ;

  case CRI_CASE_DIR:
     buff << "CASE" ;
     break ;

  case CRI_ENDCASE_DIR:
     buff << "END CASE" ;
     break ;

  case CRI_ENDPARALLEL_DIR:
     buff << "END PARALLEL" ;
     break ;

  case CRI_DOALLAUTO_DIR:
     buff << "DO ALL AUTOSCOPE" ;
     break ;

  case CRI_IVDEP_DIR:
     buff << "IVDEP" ;
     break ;

  case CRI_NOTASK_DIR:
     buff << "NOTASK" ;
     break ;

  case CRI_TASK_DIR:
     buff << "TASK" ;
     break ;

  case CRI_NOSF_DIR:
     buff << "NO SIDE EFFECTS" ;
     break ;

  case CRI_NOVECTOR_DIR:
     buff << "NO VECTOR" ;
     break ;

  case CRI_VECTOR_DIR:
     buff << "VECTOR" ;
     break ;

  case CRI_PREFERTASK_DIR:
     buff << "PREFERTASK" ;
     break ;

  case CRI_PREFERVECTOR_DIR:;
     buff << "PREFERVECTOR" ;
     break ;

  case CRI_CHUNKSIZE_DIR:
     buff << "CHUNKSIZE (" ;
     FortranUnparseExpression (buff, pllnd->node.fields.e1) ;
     buff << ")" ;
     break ;

  case CRI_NEXTSCALAR_DIR:
     buff << "NEXTSCALAR" ;
     break ;

  case CRI_UNROLL_DIR:
     buff << "UNROLL" ;
     break ;


  default :
    fprintf (stderr,"UnparseExpression -- variant %d not recognised\n",
	     pllnd->expr_variant ());
    break;
  }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |               static void FortranUnparseStatement                   |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void FortranUnparseStatement (SourceListing& src, PTR_STMT pbf)
{
  if (pbf == NULL)
    return;

  LineBuffer buff ;
  Comment *cptr = pbf->node.fields.cmnt_ptr ;

  // Firstly,  unparse any attached FULL comments

  while (cptr && cptr->type == FULL) {
    // unparse 'C' cptr->cstring '\n'
    buff.PrefixLine (0, 0, 1, 0, pbf) ;
    buff << 'C' ;
    buff.accept_raw (cptr->cstring()) ;
    buff << '\n' ;
    src.appendBuffer (buff) ;
    cptr = cptr->next ;
  }

  // Secondly, unparse the statement itself

  int tabs = computeIndentation (pbf) ;
  int slab = pbf->label.ftn ? pbf->label.ftn->index() : 0 ;
  buff.PrefixLine (slab, tabs, 0, 0, pbf) ;
  FortranUnparseLine (buff, pbf) ;
  
  // Finally, unparse the HALF-comment if present

  if (cptr && cptr->type == HALF) {
    // unparse " ! " cptr->cstring '\n'
    buff << " !" ;
    buff.accept_raw (cptr->cstring()) ;
  }

  if (pbf->stmt_variant () != CONTROL_END
      && pbf->stmt_variant () != LOGIF_NODE)
    buff << '\n' ;

  if (buff.Size() > 0)
    src.appendBuffer (buff) ;
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                 static void FortranUnparseLine                      |
 |                                                                     |
 +---------------------------------------------------------------------*/

static void FortranUnparseLine (LineBuffer& buff, PTR_STMT pbf)
{
  PTR_SYMB  s ;

  switch (pbf->stmt_variant ()) {
  case GLOBAL	:
    break;

  case PROG_HEDR	:		/* program header	*/
    buff << "program " ;
    if (pbf->node.program.prog_symb &&
	strcmp (pbf->node.program.prog_symb->ident (), (char *)"_MAIN")) {
      buff << pbf->node.program.prog_symb->ident () ;
    }
    break;

  case BLOCK_DATA	:		
    buff << "block data " ;
    if (pbf->node.program.prog_symb &&
	strcmp (pbf->node.program.prog_symb->ident (), (char *)"_BLOCK")) {
      buff << pbf->node.program.prog_symb->ident () ;
    }
    break;

  case PROC_HEDR	: 
    if (pbf->node.procedure.proc_symb->attr & RECURSIVE_BIT) 
      buff << "recursive" ;
    buff << "subroutine " ;
    buff << pbf->node.procedure.proc_symb->ident () ;
    buff << '(';
    s = pbf->node.procedure.proc_symb->node.proc_decl.in_list;
    while (s) {
      buff << s->ident () ;
      s = s->node.var_decl.next_in;
      if (s) buff << ',';
    }
    buff << ')';
    break;

  case FUNC_HEDR	:
    if (pbf->node.function.func_symb->attr & RECURSIVE_BIT) 
      buff << "recursive" ;
    if (pbf->idx) {
/* Was before fgg changed on 2000/12/05 -> to deal with real*8 in functions
      buff << ftype_name[TYPE_INDEX(pbf->node.function.func_symb->type->type_variant ())] ;
*/
      FortranUnparseType (buff, pbf->node.function.func_symb->type, 1);
      if (pbf->node.function.func_symb->type) 
      {
        FortranUnparseExpression (buff, pbf->node.function.func_symb
                                           ->type->node.fields.ranges);
      }

      buff << " function " ;
    } else
      buff << "function " ;
    buff << pbf->node.function.func_symb->ident () ;
    buff << '(';
    s = pbf->node.function.func_symb->node.proc_decl.in_list;
    while (s) {
      buff << s->ident () ;
      s = s->node.var_decl.next_in;
      if (s) buff << ',';
    }
    buff << ") " ;
    if (pbf->node.fields.e1)
    {
      buff << "result (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ")" ;
    }
    break;

  case ENTRY_STAT	:
    buff << "entry " ;
    buff << pbf->node.function.func_symb->ident () ;
    buff << '(';
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    /*
      s = pbf->node.function.func_symb->node.proc_decl.in_list;
      while (s) {
      buff << s->ident () ;
      s = s->node.var_decl.next_in;
      if (s) buff << ',';
      }
      */
    buff << ") " ;
    break;

  case INTERFACE_STMT:
  {
    PTR_SYMB s;
    char *c;
    
    buff << "interface " ;
    if ((s = (pbf->node.fields.symbol)))
    {
      c = s->ident ();
      if (*c == '.') 
      {
	buff << "operator (" ;
	buff << c ;
	buff << ")" ;
      }
      else if (*c == '=') 
      {
	buff << "assignment (" ;
	buff << "=" ;
	buff << ")" ;
      }
      else buff << c ;
    }
  }
    break;

  case MODULE_STMT:
    buff << "module " ;
    buff << pbf->node.fields.symbol->ident () ;
    break;

  case CASE_NODE:
    if (pbf->node.fields.e3)
    {
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ":" ;
    }
    buff << "select case (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    break;

  case SWITCH_NODE	:
    buff << "case (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.symbol)
      buff << pbf->node.fields.symbol->ident () ;
    break;

  case IF_NODE	:
    /* if (pbf->node.fields.e3)
       {
       FortranUnparseExpression (buff, pbf->node.fields.e3);
       buff << ":" ;
       } */
    generate_branch (buff, IF_NODE, "if ", pbf) ;
    break;

  case LOGIF_NODE	:
    buff << "if (" ;
    FortranUnparseExpression (buff, pbf->node.if_node.condition);
    buff << ") " ;
    FortranUnparseLine (buff, pbf->node.if_node.true_branch) ;
    break;

  case ELSEIF_NODE:
    generate_branch (buff, IF_NODE, "else if ", pbf) ;
    break;

  case ARITHIF_NODE:
    buff << "if (" ;
    FortranUnparseExpression (buff, pbf->node.if_node.condition);
    buff << ") " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    break;

  case WHERE_BLOCK_STMT:
    generate_branch (buff, WHERE_BLOCK_STMT, "where ", pbf) ;
    break;

  case WHERE_NODE:
    buff << "where (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ") " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << " = " ;
    FortranUnparseExpression (buff, pbf->node.fields.e3);
    break;

  case PARDO_NODE :
    gen_loop_header (buff, "parallel do ", pbf) ;
    break;

  case PDO_NODE :
    gen_loop_header (buff, "pdo ", pbf) ;
    break;

  case FOR_NODE	:
    if (pbf->node.fields.e3)
    {
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ":" ;
    }
    gen_loop_header (buff, "do ",pbf) ;
    break;

  case CDOALL_NODE	:
    gen_loop_header (buff, "cdoall ",pbf) ;
    break;

  case WHILE_NODE	:
    if (pbf->node.fields.e3)
    {
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ":" ;
    }
    buff << "do " ;
    if (pbf->node.for_node.doend) {
      char	label[7];
      
      sprintf (label,"%d ",pbf->node.for_node.doend->index ());
      buff << label ;
    }
    buff << " while (" ;
    FortranUnparseExpression (buff, pbf->node.while_node.condition);
    buff << ')';
    break;

  case ASSIGN_STAT:
    FortranUnparseExpression (buff, pbf->node.assign.l_value);
    buff << " = " ;
    FortranUnparseExpression (buff, pbf->node.assign.r_value);
    break;

  case IDENTIFY:
    buff << "identify " ;
    FortranUnparseExpression (buff, pbf->node.identify.l_value);
    buff << ' ';
    FortranUnparseExpression (buff, pbf->node.identify.r_value);
    break;

  case PRIVATE_STMT:
    buff << "private " ;
    if (pbf->node.fields.e1)
    {
      buff << ":: " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
    }
    break;

  case PUBLIC_STMT:
    buff << "public " ;
    if (pbf->node.fields.e1)
    {
      buff << ":: " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
    }
    break;

  case STRUCT_DECL:
  {
    PTR_EXPR l;
    buff << "type " ;
    
    if ((l = pbf->node.fields.e1))
    {
      buff << ", " ;
      FortranUnparseExpression (buff, l);
      buff << "::" ;
    }
    
    buff << pbf->node.fields.symbol->ident () ;
  }
    break;

  case SEQUENCE_STMT:
    buff << "sequence " ;
    break;

  case CONTAINS_STMT:
    buff << "contains " ;
    break;

  case OVERLOADED_ASSIGN_STAT:
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << "=" ;
    FortranUnparseExpression (buff, pbf->node.fields.e3);
    break;

  case OVERLOADED_PROC_STAT:
  case PROC_STAT	:
    buff << "call " ;
    buff << pbf->node.fields.symbol->ident () ;
    buff << '(';
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ')';
    break;

  case STMTFN_STAT:
  {PTR_SYMB p;
   PTR_EXPR body;
   
   body = pbf->node.fields.e1;
   p = body->node.fields.symbol;
   buff << p->ident () ;
   buff << '(';
   p=p->node.func_decl.in_list;
   while (p) {
     buff << p->ident () ;
     if ((p=p->node.var_decl.next_in)) buff << ',';
   }
   buff << ") = " ;
   FortranUnparseExpression (buff, body->node.fields.e1);
   break;
  }

  case SAVE_DECL:
    buff << "save " ;
    if (pbf->node.fields.e1)
      FortranUnparseExpression (buff, pbf->node.fields.e1);
// fgg remarked next lines on 2000/12/07
//  since it may be wrong in some implementations of fortran
//    else
//      buff << "all" ;
    break;

  case CONT_STAT:
    buff << "continue" ;
    break;

/* fgg added on 2000/12/05 */
  case ASSLAB_STAT:
    buff << "assign ";
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << " to ";
    buff << pbf->node.fields.symbol->ident () ;
    break;

  case FORMAT_STAT:
    buff << "format ";
    FortranUnparseExpression (buff, pbf->node.format.spec_string);
    break;

  case GOTO_NODE:
    buff << "goto " ;
    FortranUnparseExpression (buff, pbf->node.fields.e3);
    break;

  case ASSGOTO_NODE:
    buff << "goto " ;
    buff << pbf->node.fields.symbol->ident () ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case COMGOTO_NODE:
    buff << "goto (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ") " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    break;

  /* fgg added this on 2000/11/29 */
  case PAUSE_NODE:
    buff << "pause" ;
    break;

  case STOP_STAT:
    buff << "stop" ;
    if (pbf->node.fields.e1) {
      buff << ' ' ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
    }
    break;

  case RETURN_STAT:
    buff << "return" ;
    break;

  case OPTIONAL_STMT:
    buff << "optional :: " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case VAR_DECL:
  {
    PTR_EXPR p = pbf->node.fields.e1 ;
    in_var_decl = 1 ;
    with_kind = 0 ;

    /* PTR_TYPE q ;
       q = p->node.list.item->node.fields.symbol->type ;
       if (q->expr_variant () == T_ARRAY)
	 q = q->node.ar_decl.base_type ;
       buff << ftype_name[TYPE_INDEX(q->expr_variant ())] ;
       buff << ' ' ;
    */

    FortranUnparseExpression (buff, pbf->node.fields.e2);
    if (pbf->node.fields.e3)
    {
      buff << ", " ;
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << "::" ;
    }
    else buff << " " ;
    FortranUnparseExpression (buff, p);
    in_var_decl = 0;
    break;
  }

  case INTENT_STMT:
  {
    PTR_SYMB s;
    PTR_EXPR p = pbf->node.fields.e1;
    
    buff << "intent " ;
    s = p->node.list.item->node.fields.symbol;
    if (s->attr & IN_BIT)
      buff << "(in) :: " ;
    if (s->attr & OUT_BIT)
      buff << "(out) :: " ;
    if (s->attr & INOUT_BIT)
      buff << "(inout) :: " ;
    FortranUnparseExpression (buff, p);
    break;
  }

  case PARAM_DECL:
    buff << "parameter (" ;
    in_param = 1;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    in_param = 0;
    break;

  case DIM_STAT:
    buff << "dimension " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case ALLOCATABLE_STMT:
    buff << "allocatable :: " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case POINTER_STMT:
    buff << "pointer (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    break;

  case TARGET_STMT:
    buff << "target :: " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case ALLOCATE_STMT:
    buff << "allocate (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    if (pbf->node.fields.e2) 
    {
      buff << ", stat = " ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
    }
    buff << ")" ;
    break;

  case DEALLOCATE_STMT:
    buff << "deallocate (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    if (pbf->node.fields.e2) 
    {
      buff << ", stat = " ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
    }
    buff << ")" ;
    break;

  case NULLIFY_STMT:
    buff << "nullify (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    break;

  case MODULE_PROC_STMT:
    buff << "module procedure " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case POINTER_ASSIGN_STAT:
    buff << pbf->node.fields.symbol->ident () ;
    buff << "=> " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case CYCLE_STMT:
    buff << "cycle " ;
    buff << pbf->node.fields.symbol->ident () ;
    break;

  case EXIT_STMT:
    buff << "exit " ;
    buff << pbf->node.fields.symbol->ident () ;
    break;

  case USE_STMT:
    buff << "use " ;
    buff << pbf->node.fields.symbol->ident () ;
    if (pbf->node.fields.e1) 
    {
      buff << ", " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
    }
    break;

  case EQUI_STAT:
    buff << "equivalence " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break ;

  case DATA_DECL:
    buff << "data " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case IMPL_DECL:
    buff << "implicit " ;
    if (pbf->node.fields.e1 == NULL)
      buff << "none" ;
    else {
      in_impli = 1;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      in_impli = 0;
    }
    break;

  case EXTERN_STAT:
    buff << "external " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case INTRIN_STAT:
    buff << "intrinsic " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case PARREGION_NODE:
    buff << "parallel " ;
    if (pbf->node.fields.e1)
    {
      buff << "( " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ") " ;
    }
    break;

  case PARSECTIONS_NODE:
    buff << "parallel sections" ;
    if (pbf->node.fields.e1)
    {
      buff << "( " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ") " ;
    }
    break;

  case PSECTIONS_NODE:
    buff << "psections " ;
    if (pbf->node.fields.e1)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ") " ;
    }
    break;

  case SINGLEPROCESS_NODE:
    buff << "single process" ;
    if (pbf->node.fields.e1)
    {
      buff << "( " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ") " ;
    }
    break;

  case CRITSECTION_NODE:
    buff << "critical section" ;
    if (pbf->node.fields.e1)
    {
      buff << "( " ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ") " ;
    }
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ") " ;
    }
    break;

  case GUARDS_NODE:
    buff << "guards " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << "(" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ")" ;
    break;

  case LOCK_NODE:
    buff << "lock (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case UNLOCK_NODE:
    buff << "unlock (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case POST_NODE:
    buff << "post (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case WAIT_NODE:
    buff << "wait (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case CLEAR_NODE:
    buff << "clear (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    if (pbf->node.fields.e2)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case POSTSEQ_NODE:
    buff << "post (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ", " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ")" ;
    if (pbf->node.fields.e3)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ")" ;
    }
    break;

  case WAITSEQ_NODE:
    buff << "wait (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ", " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ")" ;
    if (pbf->node.fields.e3)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ")" ;
    }
    break;

  case SETSEQ_NODE:
    buff << "set (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ", " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ")" ;
    if (pbf->node.fields.e3)
    {
      buff << "guards (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e3);
      buff << ")" ;
    }
    break;

  case SECTION_NODE:
    buff << "section" ;
    if (pbf->node.fields.e1)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, pbf->node.fields.e1);
      buff << ")" ;
    }
    if (pbf->node.fields.e2)
    {
      buff << "wait (" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ")" ;
    }
    break;

  case ASSIGN_NODE:
    buff << "assign ( " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    break;

  case RELEASE_NODE:
    buff << "release ( " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    buff << ")" ;
    break;

  case PRIVATE_NODE:
    buff << "private " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case READ_STAT:
  {
    PTR_EXPR p;
    PTR_EXPR q;
    
    buff << "read " ;
    p = pbf->node.fields.e2;
    q = p->node.fields.e1;
    
    if ((p->expr_variant () == EXPR_LIST)  ||
	((p->expr_variant () == SPEC_PAIR) && 
	 (strcmp ((*q->node.string_val)(),"fmt") != 0)))
    {
      buff << "(" ;
      FortranUnparseExpression (buff, pbf->node.fields.e2);
      buff << ") " ;
    }
    else 
    {
      FortranUnparseExpression (buff, pbf->node.fields.e2->node.fields.e2);
      if (pbf->node.fields.e1 != 0)
	buff << ", " ;
    }
    FortranUnparseExpression (buff, pbf->node.fields.e1);
  }
    break;

  case WRITE_STAT:
    buff << "write " ;
    buff << "(" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ") " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case PRINT_STAT:
    buff << "print " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2->node.fields.e2);
    if (pbf->node.fields.e1 != 0)
      buff << ", " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case OPEN_STAT:
    buff << "open " ;
    buff << "(" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ") " ;
    break;

  case CLOSE_STAT:
    buff << "close " ;
    buff << "(" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ") " ;
    break;

  case INQUIRE_STAT:
    buff << "inquire " ;
    buff << "(" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    buff << ") " ;
    break;

  case SKIPPASTEOF_NODE:
  {
    PTR_EXPR p;
    PTR_EXPR q;
    
    buff << "skip past eof " ;
    p = pbf->node.fields.e2;
    q = p->node.fields.e2;
    
    if (p->expr_variant () == EXPR_LIST)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, p);
      buff << ") " ;
    }
    else FortranUnparseExpression (buff, q);
  }
    break;

  case BACKSPACE_STAT:
  {
    PTR_EXPR p;
    PTR_EXPR q;
    
    buff << "backspace " ;
    p = pbf->node.fields.e2;
    q = p->node.fields.e2;
    
    if (p->expr_variant () == EXPR_LIST)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, p);
      buff << ") " ;
    }
    else FortranUnparseExpression (buff, q);
  }
    break;

  case ENDFILE_STAT:
  {
    PTR_EXPR p;
    PTR_EXPR q;
    
    buff << "endfile " ;
    p = pbf->node.fields.e2;
    q = p->node.fields.e2;
    
    if (p->expr_variant () == EXPR_LIST)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, p);
      buff << ") " ;
    }
    else FortranUnparseExpression (buff, q);
  }
    break;

  case REWIND_STAT:
  {
    PTR_EXPR p;
    PTR_EXPR q;
    
    buff << "rewind " ;
    p = pbf->node.fields.e2;
    q = p->node.fields.e2;
    
    if (p->expr_variant () == EXPR_LIST)
    {
      buff << "(" ;
      FortranUnparseExpression (buff, p);
      buff << ") " ;
    }
    else FortranUnparseExpression (buff, q);
  }
    break;

  case OTHERIO_STAT:
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case COMM_STAT:
    buff << "common " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case NAMELIST_STAT:
    buff << "namelist " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break;

  case CONTROL_END:
    break;

  case COMMENT_STAT:
    buff << "C " ;
    buff.accept_raw (pbf->node.fields.cmnt_ptr->cstring()) ;
    break ;

  case INCLUDE_STAT:
    buff << "include " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break ;

  // Added for CRI Extensions
    
  case CRI_DIRECTIVE:
    buff << "!DIR$ " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    break ;


  case CRI_AUTOTASKING:
    buff << "!MIC$ " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    break ;

  // Added for Convex Fortran extensions
    
  case CVX_DIRECTIVE:
    buff << "C$DIR " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    break ;

  case CVX_OPTIONS_STMT:
    buff << "options " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    break ;

  case CVX_ACCEPT_STMT:
    {
      PTR_EXPR p = pbf->node.fields.e2 ;
      PTR_EXPR q = p ? p->node.fields.e1 : 0 ;
      
      buff << "accept " ;
      
      if (p->expr_variant () == VAR_REF)
	FortranUnparseExpression (buff, p) ;
      else 
	{
	  if ((p->expr_variant () == EXPR_LIST)  ||
	      ((p->expr_variant () == SPEC_PAIR) && 
	       (strcmp ((*q->node.string_val)(),"fmt") != 0)))
	    {
	      buff << "(" ;
	      FortranUnparseExpression (buff, pbf->node.fields.e2);
	      buff << ") " ;
	    }
	  else 
	    {
	      FortranUnparseExpression (buff, pbf->node.fields.e2->node.fields.e2);
	      if (pbf->node.fields.e1 != 0)
		buff << ", " ;
	    }
	  FortranUnparseExpression (buff, pbf->node.fields.e1);
	}
    }
    break ;

  case CVX_DECODE_STMT:
    buff << "decode (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    if (pbf->node.fields.e3)
      {
	buff << ", " ;
	FortranUnparseExpression (buff, pbf->node.fields.e3);
      }
    buff << ")" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break ;

  case CVX_ENCODE_STMT:
    buff << "encode (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e2);
    if (pbf->node.fields.e3)
      {
	buff << ", " ;
	FortranUnparseExpression (buff, pbf->node.fields.e3);
      }
    buff << ")" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1);
    break ;

  case CVX_TYPE_STMT:
    buff << "type " ;
    FortranUnparseExpression (buff, pbf->node.fields.e2->node.fields.e2) ;
    if (pbf->node.fields.e1 != 0)
      buff << ", " ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    break ;

  case CVX_FIND_STMT:
    buff << "find (" ;
    FortranUnparseExpression (buff, pbf->node.fields.e1) ;
    buff << ")" ;
    break ;

  default:
    break;
  }
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                     LineBuffer::TabIndent                           |
 |                                                                     |
 +---------------------------------------------------------------------*/

void LineBuffer::TabIndent (int n)
{
  for (int i = 0; i < n; i++) {
    (*this) << "  " ;
  }
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |               static void computeIndentation                        |
 |                                                                     |
 +---------------------------------------------------------------------*/

static int computeIndentation (PTR_STMT bf)
{
  int count = 0;
  
  while (bf && (bf->stmt_variant () != PROG_HEDR 
		&& bf->stmt_variant () != PROC_HEDR 
		&& bf->stmt_variant () != FUNC_HEDR 
		&& bf->stmt_variant () != GLOBAL))
  {
    if (bf->stmt_variant () != ELSEIF_NODE)
      count++;

    bf = bf->control_parent;
  }

  return count ;
}


/*---------------------------------------------------------------------+
 |                                                                     |
 |                   SourceListing::~SourceListing                     |
 |                                                                     |
 +---------------------------------------------------------------------*/

SourceListing::~SourceListing () {
  SourceListingIterator next_line (*this) ;
  SourceLine *line ;
  
  while (line = next_line() )
    delete line ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                   SourceListing::appendBuffer                       |
 |                                                                     |
 | Add the line, or lines, contained in buff - representing statement  |
 | stmt - to the SourceListing object invoked by the call.             |
 |                                                                     |
 | If necessary, the text in buff will be broken across multiple lines |
 | and indented and numbered as required.                              |
 |                                                                     |
 +---------------------------------------------------------------------*/

void SourceListing::appendBuffer ()
{
  appendBuffer ((LineBuffer&)(*this)) ;
}

/*---------------------------------------------------------------------+
 |                                                                     |
 |                   SourceListing::appendBuffer                       |
 |                                                                     |
 | Add the line, or lines, contained in buff - representing statement  |
 | stmt - to the SourceListing object invoked by the call.             |
 |                                                                     |
 | If necessary, the text in buff will be broken across multiple lines |
 | and indented and numbered as required.                              |
 |                                                                     |
 | If the outputFormat is HTML_TEXT, then text between and including   |
 | each pair of '<' '>' characters is not counted as part of the line. |
 |                                                                     |
 +---------------------------------------------------------------------*/

void SourceListing::appendBuffer (LineBuffer& buff)
{
  // copy lines to card[] and append them to this SourceListing object.

  if (buff.isComment || buff.isDirective || lang != F77)
    append (new SourceLine (buff, buff.stmt, buff.stmt ? buff.stmt->g_line : -1)) ;

  else {
    int p ;
    LineBuffer card ;
    char prefix [10] ;

    // output <label_prefix> <indentation> <first n characters>

    if (buff.label)
      sprintf (prefix, "%5d ", buff.label) ;
    else
      sprintf (prefix, "      ") ;

    card << prefix ;
    card.TabIndent (buff.tabs) ;

    if (ofmt == PLAIN_TEXT)
      for (p = 0; card.pos < FORTRAN_LINE_LENGTH && p < buff.pos; p++)
	card << buff.buff [p] ;
    else {
      int chars = 0, in_tag = 0, in_quote = 0, in_string = 0 ;

      // TBD: need to check that card does not overflow!

      for (p = 0; chars < FORTRAN_LINE_LENGTH && p < buff.pos; p++) {
	char ch = buff.buff [p] ;
	card << ch ;
	if (!in_tag && !in_quote)
	  if (ch == '<')
	    in_tag = 1 ;      // previously outside tag
	  else {
	    ++chars ;
	    if (ch == '\'')
	      in_quote = 1 ;  // previously outside quote
	  }
	else if (in_quote)
	  {
	    ++chars ;
	    if (ch == '\'')
	      in_quote = 0 ;
	  }
	else if (in_tag)
	  {
	    if (in_string && ch == '"')
	      in_string = 0 ;
	    else if (!in_string)
	      if (ch == '>')
		in_tag = 0 ;
	      else if (ch == '"')
		in_string = 1 ;
	  }  
      }
    }
    
    if (card.LastChar() != '\n')
      card << '\n' ;

    append (new SourceLine (card, buff.stmt, buff.stmt ? buff.stmt->g_line : -1)) ;

    // repeat <continuation_prefix> <next n characters> until done

    while (p < buff.pos)
    {
      card.pos = 0 ;
      card << "     +" ;

      // for ( ; card.pos < FORTRAN_LINE_LENGTH && p < buff.pos; p++)
      //   card << buff.buff [p] ;

      if (ofmt == PLAIN_TEXT)
	for ( ; card.pos < FORTRAN_LINE_LENGTH && p < buff.pos; p++)
	  card << buff.buff [p] ;
      else {
	int chars = 0, in_tag = 0, in_quote = 0, in_string = 0 ;
	
	// TBD: need to check that card does not overflow!
	
	for ( ; chars < FORTRAN_LINE_LENGTH && p < buff.pos; p++) {
	  char ch = buff.buff [p] ;
	  card << ch ;
	  if (!in_tag && !in_quote)
	    if (ch == '<')
	      in_tag = 1 ;      // previously outside tag
	    else {
	      ++chars ;
	      if (ch == '\'')
		in_quote = 1 ;  // previously outside quote
	    }
	  else if (in_quote)
	    {
	      ++chars ;
	      if (ch == '\'')
		in_quote = 0 ;
	    }
	  else if (in_tag)
	    {
	      if (in_string && ch == '"')
		in_string = 0 ;
	      else if (!in_string)
		if (ch == '>')
		  in_tag = 0 ;
		else if (ch == '"')
		  in_string = 1 ;
	    }  
	}
      }
      
      if (card.LastChar() != '\n')
	card << '\n' ;
      
      append (new SourceLine (card, buff.stmt, buff.stmt ? buff.stmt->g_line : -1)) ;
    }
  }
}


void CcUnparseStack (SourceListing& src, TypeStack &ts, SymbolStack& ss)
{
  PTR_TYPE e, ptr_to ;
  PTR_SYMB fs ;

  while (!(ts.is_empty()))
    {
      e = ts.pop() ;
      ptr_to = ts.pop() ;
      if (ptr_to)
	ts.push(ptr_to);
      type_tag ev = e->type_variant() ;
      switch (ev)
	{
	case T_POINTER:
	  break ;
	case T_FUNCTION:
	  if (ptr_to && ptr_to->type_variant() == T_POINTER)
	    src << ")" ;
	  src << " (" ;
	  fs = ss.pop() ;
	  fs->CcUnparseParams (src, e) ;
	  src << ")" ;
	  break ;
	case T_ARRAY:
	  if (ptr_to && ptr_to->type_variant() == T_POINTER)
	    src << ")" ;
	  src << "[" ;
	  e->node.fields.ranges->CcUnparseExpression (src, 0);
	  src << "]" ;
	  break ;
	}
    }
}
