//
// Class CSIParse for simple instruction parsing
//
// Copyright (C) 2000-2004 by Grigori G.Fursin
//
// http://homepages.inf.ed.ac.uk/gfursin
//
// ICSA, School of Informatics,
// University of Edinburgh, UK
//

#include "siparse.h"

#ifndef W32
#  include "fggunix.h"
#endif

#include <string.h> //+defines NULL
#include <stdlib.h>

#include <fstream.h>

#include <ctype.h> //for isgraph()

//****************************************************************************
//"Initialization"
//
//Input:  NumOptMax   = Max Number of Options
//        LenStrMax   = Max Length for String (Option)
//        LenParamMax = Max Length for String (Parameter)
//        IgnoreCase  = 1 - Ignore case, 0 - Don't ignore case
//
//Return: 0  - Ok
//        >0 - Error
//

int CSIParse::Init(int NumOptMax,
                   int LenStrMax,
                   int LenParamMax,
                   int IgnoreCase)
{
  int misc1, misc2;
  int x;

  XNumOptMax=NumOptMax;
  XLenStrMax=LenStrMax;
  XLenParamMax=LenParamMax;
  XIgnoreCase=IgnoreCase;

  if (NumOptMax<1) return 1;

  if (LenStrMax<2) return 1;

  if ((IgnoreCase!=0) && (IgnoreCase!=1)) return 1;

  misc1=XNumOptMax*XLenStrMax;
  misc2=XNumOptMax*XLenParamMax;
 
  Instr=new char [misc1];
  HasParam=new int[XNumOptMax];
  Sep=new char [XNumOptMax]; //separator if parameter exist
  Param=new char[misc2];

  for (x=0; x<misc1; x++) Instr[x]=0;
  for (x=0; x<misc2; x++) Param[x]=0;
  for (x=0; x<XNumOptMax; x++) 
  {
    HasParam[x]=0;
    Sep[x]=0;
  }

  CurParam=0;

  //temporal
  inst=new char[XLenStrMax];
  prm=new char[XLenParamMax];

  return 0;
}

//****************************************************************************
//"Add Instruction"
//
//Input:  Opt1     = Option 1
//        Opt2     = Option 2 (or "") - alternative option
//        Opt3     = Option 3 (or "") - alternative option
//        Default  = Default Parameter
//        HasParam = 0 - Option doesn't have parameter, 1 - it has
//
//Return: 0  - Ok
//        >0 - Error
//
int CSIParse::Add(char* Inst, int HasPrm, char Separ, char* Default)
{
  if (CurParam>XNumOptMax) return 1;

  strcpy(&Instr[XLenStrMax*CurParam], Inst);
  HasParam[CurParam]=HasPrm;
  Sep[CurParam]=Separ;
  strcpy(&Param[XLenParamMax*CurParam], Default);

  if (XIgnoreCase==1)
  {
     strcpy(&Instr[XLenStrMax*CurParam],strlwr(&Instr[XLenStrMax*CurParam]));
  }

  CurParam++;

  return 0;
}

//****************************************************************************
//"Parse Command Line"
//
//Return: <0  - Error
//        >=0 - Identificator 
//
int CSIParse::Parse(char* str)
{
  int nPos=0;
  int nLen=strlen(str);

  if (nLen==0) return -1; //Line is empty

  //ignoring first spaces
  while (str[nPos]==' ')
  {
    nPos++;
    if (nPos>=nLen) return -1; //Line is empty
  }

  //getting instruction until end/separator/' '
  inst[0]=0;
  int nPos1=0;

  int found1=0;
  while (found1==0)
  {
    if ((!islower(str[nPos])) && (!isupper(str[nPos])) && (str[nPos]!='_')) 
    {
      found1=1;
    }
    else
    {
      inst[nPos1]=str[nPos];

      nPos++;
      nPos1++;

      if (nPos>=nLen) found1=1;
    }
  }

  inst[nPos1]=0;
  if (XIgnoreCase==1) strcpy(inst, strlwr(inst));

  //Search for instruction
  found1=0;
  int i;
  for (i=0; i<CurParam; i++)
  {  
    if (strcmp(&Instr[XLenStrMax*i], inst)==0)
    {
      found1=1;
      break;
    }
  }

  if (found1==0) return -2; // Unknown instruction

  if (nPos>=nLen)
  {
    if (HasParam[i]==0) return i; //Return instruction
    return -3; //Instruction needs parameter
  }

  //ignoring spaces
  int foundspaces=0;
  while (str[nPos]==' ')
  {
    foundspaces=1;
    nPos++;
    if (nPos>=nLen) 
    {
      if (HasParam[i]==0) return i; //Return instruction
      return -3; //Instruction needs parameter
    };
  }

  if (HasParam[i]==0) return -4; //Instruction doesn't need parameters

  if (Sep[i]==' ')
  {
    if (foundspaces==0) return -5; //Wrong separator
  }
  else
  {
    if (Sep[i]!=str[nPos]) return -5; //Wrong separator
    nPos++;
  }

  if (nPos>=nLen) return -3; //Instruction needs parameter

  //ignoring spaces
  while (str[nPos]==' ')
  {
    nPos++;
    if (nPos>=nLen) return -3; //Instruction needs parameter
  }

  //getting parameter until end/' '
  prm[0]=0;
  nPos1=0;

  while (nPos<nLen)
  {
    prm[nPos1]=str[nPos];

    nPos++;
    nPos1++;
  }
  prm[nPos1]=0;

  nLen=strlen(prm);

  for (int j=nLen-1; j--; j>0)
  {
    if (prm[j]==' ') 
    {
      prm[j]=0;
    } 
    else break;
  }

  strcpy(&Param[XLenParamMax*i], prm);

  return i; //Return instruction
}

//****************************************************************************
//"Get String Parameter"
//
char* CSIParse::GetParamStr(int NumParam)
{   
  if ((NumParam<0) || (NumParam>=CurParam)) return NULL;

  return &Param[XLenParamMax*NumParam];
}

//****************************************************************************
//"Get LowerCase String Parameter"
//
char* CSIParse::GetParamStrLwr(int NumParam)
{   
  if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return NULL;

  return strlwr(&Param[XLenParamMax*NumParam]);
}

//****************************************************************************
//"Get Long Parameter"
//
long CSIParse::GetParamLong(int NumParam)
{
  if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return 0;

  return atol(&Param[XLenParamMax*NumParam]);
}

//****************************************************************************
//"Get Int Parameter"
//
int CSIParse::GetParamInt(int NumParam)
{
  if ((NumParam<0) || (NumParam>=CurParam) || (HasParam[NumParam]==0)) return 0;

  return atoi(&Param[XLenParamMax*NumParam]);
}

