/*--------------------------------------------------------------------+
 | OCTAVE - Optimising Compiler Tool And Visualisation Environment    |
 |                                                                    |
 | Copyright (C) University of Edinburgh, 1993.                       |
 |                                                                    |
 | OCTAVE is free software; you can redistribute it and/or modify     |
 | it under the terms of the GNU General Public License as published  |
 | by the Free Software Foundation; either version 1, or (at your     |
 | option) any later version.                                         |
 |                                                                    |
 | OCTAVE is distributed in the hope that it will be useful,          |
 | but WITHOUT ANY WARRANTY; without even the implied warranty of     |
 | MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      |
 | GNU General Public License for more details.                       |
 |                                                                    |
 | You should have received a copy of the GNU General Public License  |
 | along with OCTAVE; see the file COPYING.  If not, write to         |
 | the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,   |
 | USA.                                                               |
 +--------------------------------------------------------------------*/

// Source for library of meta-set operations and set classes

#if ! defined _sets_h_
#define _sets_h_

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>   // for qsort()
#include <signal.h>   // for SIGINT and raise()
#include <string.h>   // for strcat() and strlen()

#include "Basic.h"    // for basic scalar types
#include "Errors.h"   // for direct error messages
#include "Map.h"      // for class T to int mappings

// Some local macros

#define ENCODED_STATES(_x_,_y_) (((_x_).U)<<1 | ((_y_).U))
#define ELEMENT_STRING_SIZE 32
#define MAX_SUBSET_STRING (ELEMENT_STRING_SIZE * CARDINALITY)
#define MAX_SET_STRING 120000
#define MAX_SET_CONTENTS 10000
#define FATAL_ERROR(_s_) { fatal(_s_) ;}
#define CHECK_POINTER(_p_,_s_) if (!(_p_)) \
  { \
    char buf[1000] ; \
    sprintf (buf, "NULL pointer passed to '%s'", _s_) ; \
    FATAL_ERROR(buf) ; \
  }

// Operand checking is enforced in this release of Octave

#if 1
#  define CHECK_OPD(_s_) checkOperand (_s_)
#else
#  define CHECK_OPD(_s_) /* no checking */
#endif


/*--------------------------------------------------------------------+
 |                                                                    |
 |            Summary of Set class and Set operations                 |
 |                                                                    |
 | Sets are stored as an ADT table, in which the table is broken up   |
 | into finite subsets. Each subset is stored as a contiguous bit-    |
 | vector. Subsets are composed to form a full set in two ways. They  |
 | are linked together in a linear list, sorted in descending element |
 | order. The subsets are also reachable from a hash table located    |
 | centrally within the root data structure for each set. This        |
 | permits fast location of an individual subset when performing      |
 | membership tests, or when performing insertion of an element into  |
 | an existing subset. Insertion of a new subset is less frequent,    |
 | and requires a list insertion operation on the linear list of      |
 | subsets.                                                           |
 |                                                                    |
 | Operator overloading is used extensively to support set notation   |
 | as an extension to the language. The supported operators are:-     |
 |                                                                    |
 |  Operator      Operation                Syntax                     |
 |  ----------------------------------------------------------------  |
 |     =          assignment               Set = Set                  |
 |     +          union                    Set + Set                  |
 |     -          difference               Set - Set                  |
 |     &&         intersection             Set * Set                  |
 |     ==         equivalence              Set == Set                 |
 |     !=         non-equivalence          Set != Set                 |
 |     +=         add member               Set += int                 |
 |     -=         remove member            Set -= int                 |
 |     []         set membership           Set[int]                   |
 |  ----------------------------------------------------------------  |
 |     +=         union-assignment         Set += Set ;               |
 |     -=         difference-assignment    Set -= Set ;               |
 |     &=         intersection-assignment  Set &= Set ;               |
 |  ----------------------------------------------------------------  |
 |    (where 'Set' is any set expression)                             |
 |                                                                    |
 +--------------------------------------------------------------------*/

// Constants for subset type (machine dependent)

#ifdef _ILP64_
#define BITS_PER_WORD 64
#else
#define BITS_PER_WORD 32
#endif // _ILP64_

#define CARDINALITY 64
#define WORDS_PER_SET (CARDINALITY / BITS_PER_WORD)

// Constants & macros for root node type (machine independent)

#define BASE_FUNCTION(_index_) ((_index_) - ((_index_) % CARDINALITY))
#define OFFSET_FUNCTION(_index_) ((_index_) % CARDINALITY)


// Declare and/or define external global variables for profiling
// the usage of member functions defined within the SetOf and SetNode
// classes regardless of their specialization.
// These are only defined if we not compiling the libobase.a library,
// otherwise they are declared external.

#ifdef _LIBOBASE_
#  define GLOBAL_INT(_v_) int _v_ = 0 ;
#else
#  define GLOBAL_INT(_v_) extern int _v_ ;
#endif

GLOBAL_INT(_oct_nb_new_sets)
GLOBAL_INT(_oct_nb_sets_freed)
GLOBAL_INT(_oct_nb_new_nodes)
GLOBAL_INT(_oct_nb_nodes_freed)
GLOBAL_INT(_oct_nb_emptys)
GLOBAL_INT(_oct_nb_universals)
GLOBAL_INT(_oct_nb_unions)
GLOBAL_INT(_oct_nb_intersections)
GLOBAL_INT(_oct_nb_differences)
GLOBAL_INT(_oct_nb_equivalences)
GLOBAL_INT(_oct_nb_non_equivalences)
GLOBAL_INT(_oct_nb_copys)
GLOBAL_INT(_oct_nb_empty_preds)
GLOBAL_INT(_oct_nb_universal_preds)
GLOBAL_INT(_oct_nb_non_empty_preds)
GLOBAL_INT(_oct_nb_is_members)
GLOBAL_INT(_oct_nb_add_members)
GLOBAL_INT(_oct_nb_remove_members)

// Prototypes for some static helper functions...

static inline void check_subSet_index (int) ;

// Enumerated types...

       /////////////////////
       //
       // Enumerates the mode with which a Set object can
       // be created.
       //
       //  [data] SET_ACTUAL  implies the Set is a temporary
       //  [data] SET_STORED  means the Set has a static name
       //
typedef enum { SET_ACTUAL, SET_STORED } set_mode_t ;

// Define the subSet type as a vector of unsigned long...

       /////////////////////
       //
       // Storage for a dense bit-vector, used by subSet class
       //
typedef unsigned long subSet [WORDS_PER_SET] ;

/*----------- Common subSet function prototypes  -------------*/

void copy_subset (subSet, subSet) ;
void subset_union (subSet, subSet, subSet) ;
int subset_intersection (subSet, subSet, subSet) ;
int subset_difference (subSet, subSet, subSet) ;
int subset_equivalence (subSet, subSet) ;
int intcompare (const void *, const void *) ;
void make_subset (subSet, int, int) ;
char *print_compressed_Set (int, FILE*, int *) ;
char* print_compressed_set(int, FILE*, int *);
int is_invalid_ID (int) ;
int is_valid_ID (int) ;
void print_set_stats() ;

// class and template forward declarations...

class setNode ;
class Set ;
class SetIterator ;
template <class T> class SetOf ;
template <class T> class SetIteratorOf ;
template <class T> class Universe ;
template <class T> class Universe_Iterator ;
template<class T1, class T2> class Co_Universe ;



/*--------------------------------------------------------------+
 |                                                              |
 |                       class SetNode                          |
 |                                                              |
 | The SetNode class contains a subset of a complete Set, and   |
 | is stored as a bit-vector which covers CARDINALITY members   |
 | in the range (SetNode::base, SetNode::base+CARDINALITY-1).   |
 |                                                              |
 | This class defines bit-level set operations between subsets  |
 | with identical base values, and these are then used by the   |
 | Set class, and derivatives thereof, to implement a full set  |
 | algebra on disjoint sets.                                    |
 |                                                              |
 +--------------------------------------------------------------*/

class SetNode {

//   template <class T> friend class SetOf ;
//   template <class T> friend class SetIteratorOf ;

public:

  SetNode (int bs = 0)
  { 
    ++_oct_nb_new_nodes ;
    base = bs ;
    thread = 0 ;
    for (int i = 0; i < WORDS_PER_SET; i++)
      items[i] = 0 ;
  }

  SetNode (SetNode& sn)
  { 
    ++_oct_nb_new_nodes ;
    base = sn.base ;
    thread = 0 ;
    for (int i = 0; i < WORDS_PER_SET; i++)
      items[i] = sn.items[i] ;
  }

  ~SetNode () { ++_oct_nb_nodes_freed ; }

  int index () { return base ; }
  int variant () { return 0 ; }

  void operator+= (int i)
  {
    check_subSet_index (i) ;
    this->items[i/BITS_PER_WORD] |= 1 << (i % BITS_PER_WORD) ;
  }

  bool operator[] (int i)
  {
    return (bool)((items[i/BITS_PER_WORD] >> (i % BITS_PER_WORD)) & 1) ;
  }
  
  bool operator-= (int i)
  {
    check_subSet_index (i) ;
    if (!(items[i/BITS_PER_WORD] &= (1 << (i % BITS_PER_WORD)) ^ -1)) {
      // this word of the subset is empty, so check for a completely
      // empty subset...
      return is_empty_subset() ;
    }
    else
      return false ;
  }

  bool is_empty_subset ()
  {
    for (int i = 0; i < WORDS_PER_SET; i++)
      if (items[i])
	return false ;
    return true ;
  }

  bool is_full_subset ()
  {
    for (int i = 0; i < WORDS_PER_SET; i++)
      if (items[i] ^ (unsigned long)-1)
	return false ;
    return true ;
  }

  // private:

  int base ;       // node stores members 'base' to 'base+CARDINALITY-1
  SetNode *thread ;// thread used for linear ordering of set nodes 
  subSet items ;   // subset storage for CARDINALITY contiguous members

  char *print (int, FILE*, int*) ;
  void free_thread () ;
  char *print_subset (int, FILE*, int, int*) ;
};



// CLASS GROUP: classes for the Set container

struct FiniteInt {
  int ival ;
  int index() { return ival ; }
  int variant() { return 0 ; }
  void setIndex (int i) { ival = i ; }
  FiniteInt (int i) { ival = i; }
};


/*--------------------------------------------------------------

CLASS
   Universe

   A template class which contains particular universes over which
   sets of class T objects can be constructed.                                              
           
DESCRIPTION
                                                     
   Objects of class T can only be members of the SetOf&ltT&gt class 
   if they satisfy the following conditions.                    
                              
   <dl>                                  
     <dt><img src="ball.gif">
          class T must declare Universe&ltT&gt as a friend class    
                                                                
     <dt><img src="ball.gif">
          class T must have an integer data member 'id'.        
                                                                
     <dt><img src="ball.gif">
          class T must provide a public int member function     
          called index(), which returns the value of 'id'.      
                                                                
     <dt><img src="ball.gif">
          new objects of class T must be created by either :-   

       <dl>
          <dt><img src="ball.gif">
               member function Universe&ltT&gt::make_member(),      
               which returns a T* pointing to the newly-created 
               object of class T <strong>or</strong>                                                       
          <dt><img src="ball.gif">
               member function Universe&ltT&gt::make_member(int i), 
               which returns a T* pointing to the newly-created 
               object of class T. The 'id' field of the new     
               object is set to 'i', provided this value is not 
               already in use by another member of this         
               universe <strong>or</strong>                                                       

          <dt><img src="ball.gif">
               member function Universe&ltT&gt::add_member(T* m),   
               which adds the object pointed to by 'm' to the   
               universe, and sets it 'id' field to the next     
               unique identifier for this universe.             
          </dl>                                                                
                                                                
     <dt><img src="ball.gif">
          class T must have a constructor function with         
          prototype as shown below, performing at least the     
          action of initialising the 'id' field of the object.  

          EXAMPLE
                T::T (int i) { id = i ; }                       
          END
    </dl>

---------------------------------------------------------------*/

template<class T> class Universe {

friend class SetOf<T> ;
friend class Universe_Iterator<T> ;

public:

	/////////////////////////////////
	//
	//
	//
	//
	//
  int lwb() { return l ; }

	/////////////////////////////////
	//
	//
	//
	//
	//
  int upb() { return u ; }

	/////////////////////////////////
	//
	//
	//
	//
	//
  T* create (int i) { return new T (i) ; }

	/////////////////////////////////
	//
	//
	//
	//
	//
  T* make_member()
  {
    T* item = create (++u) ;
    if (l > u) l = u ;
    domain += item ;
    touch () ;
    return item ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  T* make_member(int i)
  {
    touch () ;
    if (domain[i])
      return domain[i] ;
    else 
    {
      T* item = create (i) ;
      if (i > u) u = i ;
      if (l > u) l = u ;
      domain += item ;
      return item ;
    }
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  void add_member (T* m)
  {
    touch () ;
    m->setIndex (++u) ;
    if (l > u) l = u ;
    domain += m ;
  }

	/////////////////////////////////
	//
	// Inserts an existing object of class T*
	// in this Universe if it is not already present.
	//
	// [in] T*  a pointer to the new member
	// [in] int the new member's index
	//
  void include_member (T* m)
  {
    if (!m) return ;
    int i = m->index() ;
    if (!domain[i]) {
      touch () ;
      domain += m ;
      if (l > i) l = i ;
      if (u < i) u = i ;
    }
  }

	/////////////////////////////////
	//
	// Inserts an existing object of class T&
	// in this Universe.
	//
	// [in] T&  a reference to the new member
	// [in] int the new member's index
	//
  void include_member (T& m)
  {
    int i = m.index() ;
    if (!domain[i]) {
      touch () ;
      domain += m ;
      if (l > i) l = i ;
      if (u < i) u = i ;
    }
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  T* get_member(int i)
  {
    return domain[i] ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  void touch () { ++ vstamp ; }

	/////////////////////////////////
	//
	//
	//
	//
	//
  Universe (int low = INT_MAX, int high = 0)
  { 
    l = low ; 
    u = high ;
    vstamp = 1 ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  ~Universe() { domain.zap() ; }

private:

  Map<T> domain ;  // the mapping [int => T*]
  int l, u ;       // lower and upper bounds on integer domain
  int vstamp ;     // incremented each time universe changes
};

template <class T> class Universe_Iterator : public Map_Iterator<T> {
public:
  Universe_Iterator (Universe<T>& u) : Map_Iterator<T> (u.domain) {}
};

/*--------------------------------------------------------------

CLASS
   SetOf
                                                                
   The SetOf template class permits the storage and manipulation of
   sets of objects.
     
DESCRIPTION
                                                           
   A full range of set operations is provided, and these are    
   implemented efficiently.                                     
                                                                
   Each SetOf object has a defined universe, or range of values   
   which its members can take. The lower bound and upper bound  
   on permissible member values are returned by the low() and   
   and high() functions respectively.                        
   The SetOf class is parameterised by class T, and implements
   sets of objects of class T.

   Class T must define the following public member function:
   
   EXAMPLE
         int T::index()  // returns the index of `this`
   END
                                                                
   Each SetOf is comprises a list of SetNode objects, sorted for  
   efficient O(n) combination of pairs of SetOf objects (by the Union,   
   Intersection, and Difference operators) - where n is the     
   number of elements in a set.                                 
                                                                
   Each SetOf also has a dynamic hash table (a SetOf object is derived     
   privately from the Map<SetNode> class), which permits access 
   to any element of the set in O(1) time using O(n) space.     
   Thus, a membership test is O(1), and insertion of a new      
   set member is usually O(1), though the worst case is O(n).   
                                                                
--------------------------------------------------------------*/

template<class T> class SetOf : private Map<SetNode> {

friend class SetIteratorOf<T> ;
friend class Universe<T> ;

public:

	//////////////////
	//
	//
	//
	//
  SetOf<T>& theSet()   { return *this ; }

	//////////////////
	//
	//
	//
	//
  void operator=  (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  SetOf<T>& operator+  (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  SetOf<T>& operator-  (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  SetOf<T>& operator&& (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  bool operator== (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  bool operator!= (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  void operator+= (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  void operator-= (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  void operator+= (int)  ;

	//////////////////
	//
	//
	//
	//
  void operator-= (int)  ;

	//////////////////
	//
	//
	//
	//
  bool operator[] (int)  ;

	/////////////////////////////////
	//
	//
	//
	//
	//
  bool operator[] (T& elem)
  {
    return operator[] (elem.index()) ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  bool operator[] (T* elem)
  {
    return elem ? operator[] (elem->index()) : false ;
  }
  
	/////////////////////////////////
	//
	//
	//
	//
	//
  void operator+= (T& elem)
  {
    if (u) u->include_member (elem) ;
    operator+= (elem.index()) ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  void operator+= (T* elem)
  {
    if (u) u->include_member (elem) ;
    operator+= (elem->index()) ;
  }
  
	/////////////////////////////////
	//
	//
	//
	//
	//
  void operator-= (T& elem)
  {
    if (u) u->include_member (elem) ;
    operator-=(elem.index()) ; 
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  void operator-= (T* elem)
  {
    if (u) u->include_member (elem) ;
    operator-=(elem->index()) ; 
  }

	//////////////////
	//
	//
	//
	//
  void operator&= (SetOf<T>&) ;

	//////////////////
	//
	//
	//
	//
  SetOf<T>& complement () ;

	/////////////////////////////////
	//
	//
	//
	//
	//
  SetOf<T>& operator~ ()
  {
    return this->SetOf<T>::complement () ;
  }

	//////////////////
	//
	//
	//
	//
  void make_universal () ;

	//////////////////
	//
	//
	//
	//
  void empty () ;

	//////////////////
	//
	//
	//
	//
  int lwb()  { return u ? u->l : 0 ; }

	//////////////////
	//
	//
	//
	//
  int upb() { return u ? u->u : 0 ; }

	//////////////////
	//
	//
	//
	//
  void printMap () 
  {
    Map<SetNode>::print () ;
  }

	//////////////////
	//
	//
	//
	//
  bool is_universal_set () ;

	//////////////////
	//
	//
	//
	//
  bool is_empty () ;

	//////////////////
	//
	//
	//
	//
  bool is_complemented () { return U ; }

	//////////////////
	//
	//
	//
	//
  char *print (int, FILE*) ;

	//////////////////
	//
	//
	//
	//
  int *compress () ;
  
	//////////////////
	//
	//
	//
	//
  int new_sets()    { return _oct_nb_new_sets ; }

	//////////////////
	//
	//
	//
	//
  int sets_freed()  { return _oct_nb_sets_freed ; }

	//////////////////
	//
	//
	//
	//
  int new_nodes()   { return _oct_nb_new_nodes ; }

	//////////////////
	//
	//
	//
	//
  int nodes_freed() { return _oct_nb_nodes_freed ; }

	//////////////////
	//
	//
	//
	//
  SetOf (int i = 0) : Map<SetNode> ()
  {
    id = i ;
    ++_oct_nb_new_sets ;
    temp = false ;
    ownsUniv = false ;
    U = false ;
    thread = 0 ;
    u = new Universe<T>() ;
    vstamp = u->vstamp ;
  }

	//////////////////
	//
	//
	//
	//
  SetOf (Universe<T>* uv, bool tmp = false, int i = 0) : Map<SetNode> ()
  {
    id = i ;
    ++_oct_nb_new_sets ;
    temp = tmp ;
    U = ownsUniv = false ;
    thread = 0 ;
    u = uv ;
    vstamp = uv ? uv->vstamp : 0 ;
  }

	//////////////////
	//
	//
	//
	//
  SetOf (SetOf<T>& s) : Map<SetNode> ()
  {
    ++_oct_nb_new_sets ;
    temp = true ;
    thread = 0 ;
    ownsUniv = false ;
    u = s.u ;
    vstamp = s.u ? s.u->vstamp : 0 ;
    (*this) = s ;        // uses assignment operator on Sets
  }
  
	//////////////////
	//
	//
	//
	//
  ~SetOf ()
  { 
    ++_oct_nb_sets_freed ;
    if (!temp)
      this->empty() ; 
    if (ownsUniv && u)
      delete u ;
  }

	//////////////////
	//
	//
	//
	//
  void cherish ()
  { 
    temp = false ; 
  }

	//////////////////
	//
	//
	//
	//
  void cleanup ()
  { 
    if (!temp)
      return ;
    else
      delete this ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  void set_universe (Universe<T>* uv)
  { 
    u = uv ;
    vstamp = u ? u->vstamp : 0 ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  Universe<T>* universe() { return u ; }

	//////////////////
	//
	//
	//
	//
  bool isTemp() { return temp ; }

  void setIndex (int i) { id = i ; }
  int index() { return id ; }
  int variant() { return id ; }

private:

  bool temp ;       // 1 => temporary set, 0 => this is a set variable
  bool U ;          // U != 0 => this is a complementary set
  bool ownsUniv ;   // true if this set owns its universe
  int  vstamp ;     // version stamp for universe-consistency
  SetNode* thread ; // pointer to subset with highest base value
  Universe<T>* u ;  // pointer to this set's Universe object
  int id ;          // identifier for this set, to enable sets of sets

  // inlined private member functions for Set

  void checkOperand (SetOf<T>& s)
  {
    if (s.u != u)
      FATAL_ERROR("set operands are from different universes")
    else if (vstamp != u->vstamp || s.vstamp != u->vstamp) {
      FATAL_ERROR("universe has grown since set operands were created")
      printf ("\t v1 = %d, v2 = %d, u = %d\n", vstamp, s.vstamp, u->vstamp) ;
    }
  }

  void clear_thread()
  { 
    if (thread) 
    {
      thread->free_thread() ; 
      thread = 0 ;
    }
  }

  void insert_subset (SetNode* sub)
  {
    SetNode **p = &thread ;
    while (*p && ((*p)->base < sub->base))
      p = &((*p)->thread) ;
    sub->thread = *p ;
    *p = sub ;
  }

  void install_subset(SetNode* sub)
  { 
    Map<SetNode>::operator+= (sub) ;
  }
  
  void delete_subset (SetNode* l)
  {
    SetNode **t ;
    for (t = &(this->thread) ;
	 *t ;
	 t = &((*t)->thread))
      if ((*t)->base == l->base)
	break ;

    if (*t)
    {
      *t = (*t)->thread ;
      Map<SetNode>::remove (l) ;
      delete l ;
    }
  }

  void theIntersection (SetOf<T>&, SetOf<T>&) ;
  void theUnion (SetOf<T>&, SetOf<T>&) ;
  void theDifference (SetOf<T>&, SetOf<T>&) ;
  bool theEquivalence(SetOf<T>&) ;
  bool fullEquivalence(SetOf<T>&) ;
};


/*--------------------------------------------------------------

CLASS
   SetIteratorOf

   Declares structures and member functions   
   for permitting easy iteration through the members of a set.  
    
DESCRIPTION
                                                            
   A convenience class for writing code which iterates sequentially
   through all members present in a given SetOf&ltT&gt class.                  
                                                                
   One can iterate through all members of SetOf&ltT&gt s1 as follows:

   EXAMPLE
     SetIteratorOf&ltT&gt next_T (s1) ;                               
     SetOf&ltT&gt *t ;                                                
                                                                
     while ( t = next_T() )                                       
     {                                                            
      // do something with t ;                                
     }
   END

   If you want to iterate through the actual members of a set,  
   irrespective of whether the set is stored in positive or     
   complemented form, then initialize your iterator as follows: 

   EXAMPLE                                                                
     SetIteratorOf&ltT&gt next(setvariable, SET_ACTUAL) ;                
   END
                                                             
   If you want to iterate through the stored elements of the    
   Set (eg the print function needs to do this) then initialize 
   your iterator using:-                                        

   EXAMPLE                                                             
     SetIteratorOf&ltT&gt next(setvariable, SET_STORED) ;                
   END
                                                             
   Note that SET_STORED and SET_ACTUAL are of type set_mode_t   
                                                                
----------------------------------------------------------------*/

template<class T> class SetIteratorOf {

public:

	//////////////////
	//
	//
	//
	//
  void operator++()
  {
    if ((mode == SET_ACTUAL) && complemented)
    { 
      do
      {
	++offset ;
	done = (node) ? (bool)((offset + node->base) > limit) :
	(bool)(offset > limit); 
	
	if ((offset == CARDINALITY) && node && !done) {
	  if (node->thread) {
	    offset = 0 ;
	    node = node->thread ;
	  } else {
	    offset = node->base + offset ;
	    node = 0 ;
	  }
	}
      } while ( !done && node && (*node)[offset]) ;
    }
    else
    {
      do
      {
	++offset ;
	if (offset == CARDINALITY) {
	  offset = 0 ;
	  node = node->thread ;
	  done = (bool)(!node) ;
	}
      } while ( node && !(*node)[offset] ) ;
    }
    return ;
  }
  
	//////////////////
	//
	//
	//
	//
  int operator() ()
  {
    if (done)
      return 0 ;
    if (node)
      return node->base + offset ;
    else
      return offset ;
  }

	//////////////////
	//
	//
	//
	//
  T* operator[] (int i)
  {
    if (!i || !set || !(set->u))
      return 0 ;
    T* r = set->u->get_member(i) ;
    if (!r)
      r = set->u->make_member(i) ;
    return r ;
  }
	//////////////////
	//
	//
	//
	//
  SetIteratorOf (SetOf<T>& so, set_mode_t m = SET_ACTUAL)
  { 
    done = false ;
    set = &so ;
    node = so.thread ;
    limit = set->upb() ;
    complemented = set->is_complemented() ;
    mode = m ;

    if ((mode == SET_ACTUAL) && complemented)
    {
      while (node && node->is_full_subset())
	node = node->thread ;

      if (node)
	offset = set->lwb() % CARDINALITY ;
      else
	offset = set->lwb() ;

      // get first index NOT contained in node
      while ( node && (offset < CARDINALITY) && (*node)[offset] )
	++offset ;
    }
    else
    {
      // get first index contained in node
      offset = 0 ;
      while ( node && (offset < CARDINALITY) && !(*node)[offset] )
	++offset ;
      if (!node) done = true ;
    }
  }
  
	//////////////////
	//
	//
	//
	//
  ~SetIteratorOf () { }

private:

  SetOf<T>*  set ;
  SetNode*   node ;
  int        offset ;
  int        limit ;
  bool       done ;
  bool       complemented ;
  set_mode_t mode ;
};


/*--------------------------------------------------------------
                                                                
CLASS
   Co_Universe
                            
   A template class for associating members of the Universe defined
   by class T1 with objects of the class T2.
                                       
DESCRIPTION

   The Co_Universe&ltT1, T2&gt class is a template class which      
   supports the association of members of class T1 (which must  
   be a Universe&ltT&gt class) with members of a co-universe,       
   defined as Universe&ltT2&gt.                                     
                                                                
   Thus, for example, if 'tu' is a TupleUniverse, we declare:   

    EXAMPLE                                                                
      Co_Universe<TupleUniverse, SetOfTuples> tupleDomain (tu) ;  
    END

   Then, given tuple t which is a member of tu, we can find or  
   create a mapping from tuple t to a SetOfTuples, using:-      

    EXAMPLE                                                                
      SetOfTuples* s = tupleDomain.get_codomain (t) ;             
    END
                                                              
   This finds and returns an existing mapping with t as the     
   domain element within tupleDomain, and creates such a        
   mapping (and a new SetOfTuples) if no mapping exists.        
                                                                
   Objects of class T1 and T2 can only be parameters of the     
   Co_Universe class if they satisfy the following conditions.  
                                                                
   <dl>
     <dt><img src="ball.gif">
          each parameter class P must declare Universe&ltP&gt as a  
          friend class                                          
                                                                
     <dt><img src="ball.gif">
          each parameter class P must have an integer data      
          member 'id'.                                          
                                                                
     <dt><img src="ball.gif">
          class P must provide a public int member function     
          called index(), which returns the value of 'id'.      
                                                                
     <dt><img src="ball.gif">
          new objects of class T1 must be created by either :-  

        <dl>                                                                
          <dt><img src="ball.gif">
               member function Universe&ltT1&gt::make_member(),     
               which returns T1* pointing to the newly-created  
               object of class T1.                              
	       <strong>OR</strong>

          <dt><img src="ball.gif">
               member function Universe&ltT1&gt::make_member(int i) 
               which returns a T1* pointing to the newly-created
               object of class T1. The 'id' field of the new    
               object is set to 'i', provided this value is not 
               already in use by another member of this         
               universe.
	       <strong>OR</strong>

          <dt><img src="ball.gif">
               member function Universe&ltT1&gt::add_member(T1* m), 
               which adds the object pointed to by 'm' to the   
               universe, and sets it 'id' field to the next     
               unique identifier for this universe.             
        </dl>

     <dt><img src="ball.gif">
          class T1 must have a constructor function with        
          prototype as shown below, performing at least the     
          action of initialising the 'id' field of the object.  
          EXAMPLE
                T1::T1 (int i) { id = i ; }
          END
                                                                
     <dt><img src="ball.gif">
          new objects of class T2 must be created by using the  
          Co_Universe&ltT1,T2&gt function make_associate(T1 m)      
          which checks if `m` already has an associated value   
          in this co-universe (in which case it returns a       
          pointer to it), and if not, it creates a new object   
          of class T2 and sets the id of that object to be the  
          same as the id of the domain object `m`.              
                                                                
     <dt><img src="ball.gif">
          class T2 must define a constructor function of the    
          form :-                                               
         
	  EXAMPLE
            T2::T2 (int i) { id = i ; }                         
          END
   </dl>

   Given a T1 object 't', a co-universe 'cu' which was created  
   from the universe containing 't', and a co-universe object   
   pointer 'co', the associated T2 object can be retrieved by:  
   
   EXAMPLE
     co = cu [t] ;                                              
   END

   If there is no mapping from t to a T2 object in cu, then the 
   [] operator returns a zero pointer.                          
                                                                
---------------------------------------------------------------*/

template<class T1, class T2> class Co_Universe : public Universe<T2> {

public:

	/////////////////////////////////
	//
	//
	//
	//
	//
  Co_Universe () : Universe<T2> ()    {}

	/////////////////////////////////
	//
	//
	//
	//
	//
  Co_Universe (Universe<T1> &u1) : Universe<T2> (u1.lwb(), u1.upb())    {}

	/////////////////////////////////
	//
	//
	//
	//
	//
  Co_Universe (Universe<T1> *u1) : Universe<T2> (u1->lwb(), u1->upb()) {}

	/////////////////////////////////
	//
	//
	//
	//
	//
  T2* get_codomain (T1& m)
  {
    int i = m.index() ;
    T2* c = get_member (i) ;

    if (c)
      return c ;
    else {
      return make_member (i) ;
    }
  } ;

	/////////////////////////////////
	//
	//
	//
	//
	//
  T2* get_codomain (T1* m)
  {
    if (!m)
      return 0 ;

    int i = m->index() ;

    T2* c = get_member (i) ;

    if (c)
      return c ;
    else {
      return make_member (i) ;
    }
  } ;

	/////////////////////////////////
	//
	//
	//
	//
	//
  T2* operator[] (T1& d) 
  { 
    return Universe<T2>::get_member (d.index()) ;
  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  T2* operator[] (T1* d)
  { 
    if (d)
      return Universe<T2>::get_member (d->index()) ; 
    else
      return 0 ;
  }

};


/*--------------------------------------------------------------

CLASS
   Set

   A specialization of the SetOf class for implementing simple sets
   of positive integers.

DESCRIPTION


----------------------------------------------------------------*/

class Set : public SetOf<FiniteInt> {
public:
friend class SetIterator ;

  Set () : SetOf<FiniteInt> () { }
  Set (int l, int u) : SetOf<FiniteInt> (new Universe<FiniteInt>(l,u)) { }
  Set (Universe<FiniteInt>* uv) : SetOf<FiniteInt> (uv) { }
  void operator= (SetOf<FiniteInt>& s) { SetOf<FiniteInt>::operator=(s); }
};


/*--------------------------------------------------------------
                                                                
CLASS
   SetIterator

DESCRIPTION
   A specialization of the SetIteratorOf template class for use
   with the Set specialiazation of the SetOf class.
----------------------------------------------------------------*/

class SetIterator : public SetIteratorOf<FiniteInt> {

public:

	/////////////////////////////////
	//
	//
	//
	//
	//
  SetIterator (Set& so, set_mode_t m = SET_ACTUAL) 
               : SetIteratorOf<FiniteInt> (so, m)
  {  }

	/////////////////////////////////
	//
	//
	//
	//
	//
  ~SetIterator() { }

};


// Some static helper functions - expect these to be inlined
// by -O optimizations for maximum efficiency.

/*--------------------------------------------------------------
                                                                
            static inline void check_subSet_index (int i)       
                                                                
   This routine checks that integer i is in the subset range    
   (0, CARDINALITY-1). If it is not, we flag an error.          
----------------------------------------------------------------*/

static inline void check_subSet_index (int i) 
{
  if ((i < 0) || (i >= CARDINALITY)) 
    { 
      fprintf (stderr, "Set index %d is out of bounds\n",i) ;
      raise (SIGINT) ;
    }
}

/*---------------- Template class member functions -----------------*/

/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::empty()
 *
 * Inputs:  'this' is a pointer to a SetOf variable.
 *
 * Outputs: Always returns void.
 *
 * Actions: Frees the subsets associated with this, but creates an
 *          empty root struct if none exists. Also initialises the
 *          hash table for the root set if one exists already.
 * 
 *-------------------------------------------------------------------*/

template<class T> void SetOf<T>::empty()
{
  ++_oct_nb_emptys ;
  
  CHECK_POINTER(this, "empty()") ;

  if (this) {
    this->clear() ;
    this->clear_thread() ;
    this->U = false ;
    this->vstamp = u ? u->vstamp : 0 ;
  }
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator= (SetOf<T>& s1) 
 *
 * Inputs:  's1' is an input set, which gets assigned to 'this'.
 *
 * Outputs: void output, but modifies `this`
 *
 * Actions: retains s1 (non-destructively) copying it over to `this`
 * 
 *-------------------------------------------------------------------*/

template<class T> void SetOf<T>::operator= (SetOf<T>& s1)
{
  SetNode **p0, *p1;

  ++_oct_nb_copys ;

  if (!this) return ;

  if (s1.is_empty())
    this->empty() ;
  else {
    this->U = s1.U ;
    this->vstamp = s1.vstamp ;
    if (this->thread)      // if this set had previous contents...
      this->clear () ;     // ...clear the hash table to erase them
    
    p0 = &(this->thread) ;
    for (p1 = s1.thread; p1; p1 = p1->thread) {
      if (!(*p0)) {
	*p0 = new SetNode ;
      }
      (*p0)->base = p1->base ;
      copy_subset (p1->items, (*p0)->items) ;
      this->install_subset (*p0) ;
      p0 = &((*p0)->thread) ;
    }
    (*p0)->free_thread() ;
    *p0 = 0 ;
  }
  s1.cleanup() ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: SetOf<T>& SetOf<T>::operator&& (SetOf<T>& s2) ;
 *          
 * Inputs:  '*this' and 's2' are input set objects.
 *
 * Outputs: returns the set union of the two input sets, which may
 *          create a temporary set if *this is not already a
 *          temporary object.
 *
 * Actions: *this and s2 are scanned linearly, using their threads,
 *          to form the intersection. If s2 is a temporary it is
 *          deleted.
 * 
 *-------------------------------------------------------------------*/

template<class T> SetOf<T>& SetOf<T>::operator&& (SetOf<T>& s2)
{
  ++_oct_nb_intersections ;
  CHECK_OPD(s2) ;

  // Axiom: {} Inter X = {}, and X Inter {} = {}

  if (this->is_empty()) {
    s2.cleanup() ;
    return *this ;
  }
  if (s2.is_empty()) {
    s2.cleanup() ;
    // now return an empty set
    if (temp) {
      // if this is temporary, just clear it and return it...
      empty() ;
      return *this ;
    }
    else {
      // if this is not temporary, return a new temporary (empty) set...
      SetOf<T>* e = new SetOf<T> (this->u, true) ;
      return *e ;
    }
  }

  // If this is a temporary we can re-use it, otherwise make 
  // a new temporary set.

  SetOf<T>* dst = this->temp ? this : new SetOf<T> (*this) ;
  SetOf<T>& s0 = *dst ;
  s0.temp = true ;

  switch (ENCODED_STATES(s2,*this))
  {
  case 0:  // s2 positive, this positive
    this->theIntersection(s0, s2) ;
    s0.U = false ;
    break ;
  case 1:  // s2 positive, s1 negative
    s2.theDifference (s0, *this) ;
    s0.U = false ;
    break ;
  case 2:  // s2 negative, s1 positive
    this->theDifference (s0, s2) ;
    s0.U = false ;
    break ;
  case 3:  // s2 negative, s1 negative
    s2.theUnion (s0, *this) ;
    s0.U = true ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for univeral set") ;
    break ;
  }

  // retire the input operands and return the result
  
  s2.cleanup() ;
  return s0 ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: SetOf<T>& SetOf<T>::operator+ (SetOf<T>& s2) 
 *
 * Inputs:  '*this' and 's2' are input sets. 
 *
 * Outputs: returns a the set union of the two inputs, which may
 *          create a temporary set, if the object *this is not
 *          already a temporary.
 *
 * Actions: this and s2 are scanned linearly, using their threads,
 *          to form the union between s1 and s2. 
 *          If s2 is a temporary it is deleted.
 *-------------------------------------------------------------------*/

template <class T> SetOf<T>& SetOf<T>::operator+ (SetOf<T>& s2)
{  
  ++_oct_nb_unions ;
  CHECK_OPD(s2) ;

  // Axioms: X u {} = X, and {} u X = X

  if (s2.is_empty()) {
    return *this ;
  }
  if (this->is_empty()) {
    return s2 ;
  }

  // If this is not a temporary set we neet to create one, and note that
  // it is a temporary object (until assigned by Set::operator=)

  SetOf<T>* dst = this->temp ? this : new SetOf<T> (*this) ;
  SetOf<T>& s0 = *dst ;
  s0.temp = true ;

  switch (ENCODED_STATES(*this,s2))
  {
  case 0:  // this positive, s2 positive
    this->theUnion (s0, s2) ;
    s0.U = false ;
    break ;
  case 1:  // this positive, s2 negative
    s2.theDifference (s0, *this) ;
    s0.U = true ;
    break ;
  case 2:  // this negative, s2 positive
    this->theDifference (s0, s2) ;
    s0.U = true ;
    break ;
  case 3:  // this negative, s2 negative
    this->theIntersection (s0, s2) ;
    s0.U = true ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for univeral set") ;
    break ;
  }
  
  // Discard the right-hand input operand if it is a temporary
  
  s2.cleanup() ;

  // Return the result

  return s0 ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: SetOf<T>& SetOf<T>::operator- (SetOf<T>& s2)
 *
 * Inputs:  'this' and 's2' are input sets, 
 *
 * Outputs: returns a SetOf<T>& which is the set difference of `this` and 
 *          s2, i.e. `this` with all elements in s2 removed.
 *
 * Actions: first `this` and `s2` are scanned linearly to form the
 *          difference between `this` and `s2`.
 *
 *-------------------------------------------------------------------*/

// NEEDS BRINGING UP TO DATE, SIMILAR TO + OPERATOR... COME BACK TO THIS!!!

template <class T> SetOf<T>& SetOf<T>::operator- (SetOf<T>& s2)
{
  SetOf<T>& s0 = *(new SetOf<T>());
  s0.temp = true ;

  ++_oct_nb_differences ;
  CHECK_OPD(s2) ;

  // Axiom: this - {} = this

  if (s2.is_empty()) {
    s0 = *this ;
    return s0 ;
  }
  
  // Axiom: {} - x = {}

  if (this->is_empty())
    return s0;

  switch (ENCODED_STATES(*this,s2))
  {
  case 0:  // this positive, s2 positive */
    this->theDifference (s0, s2) ;
    s0.U = false ;
    break ;
  case 1:  // this positive, s2 negative */
    this->theIntersection (s0, s2) ;
    s0.U = false ;
    break ;
  case 2:  // this negative, s2 positive */
    this->theUnion (s0, s2) ;
    s0.U = true ;
    break ;
  case 3:  // this negative, s2 negative */
    s2.theDifference (s0, *this) ;
    s0.U = false ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for complemented set") ;
    break ;
  }

  // retire the input operands and return the result

  s2.cleanup() ;
  return s0 ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator&= (SetOf<T>& s2) ;
 *          
 * Inputs:  'this' and 's2' are input set objects.
 *
 * Outputs: overwrites 'this' with the intersection of 'this' and s2
 *
 * Actions: 'this' and 's2' are scanned linearly, using their threads,
 *          to form the intersection between 'this' and 's2'. 
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::operator&= (SetOf<T>& s2)
{
  ++_oct_nb_intersections ;
  CHECK_OPD(s2) ;

  // Axiom: {} Inter X = {}, and X Inter {} = {}

  if (!this || this->is_empty())
    return ;

  if (s2.is_empty()) {
    this->empty() ;
    return ;
  }
  
  switch (ENCODED_STATES(s2,*this))
  {
  case 0:  // s2 positive, this positive
    this->theIntersection(*this, s2) ;
    break ;
  case 1:  // s2 positive, this negative
    s2.theDifference (*this, *this) ;
    this->U = false ;
    break ;
  case 2:  // s2 negative, this positive
    this->theDifference (*this, s2) ;
    break ;
  case 3:  // s2 negative, this negative
    s2.theUnion (*this, *this) ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for univeral set") ;
    break ;
  }

  // retire the input operand and exit
  
  s2.cleanup() ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator+= (SetOf<T>& s2) 
 *
 * Inputs:  '*this' and 's2' are input sets. 
 *
 * Outputs: overwrites 'this' with the set
 *          union of this and s2.
 *
 * Actions: this and s2 are scanned linearly, using their threads,
 *          to form the union between this and s2. 
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::operator+= (SetOf<T>& s2)
{
  ++_oct_nb_unions ;
  CHECK_OPD(s2) ;

  // Axioms: X u {} = X, and {} u X = X

  if (!this || s2.is_empty())
    return ;

  if (this->is_empty()) {
    *this = s2 ;
    return ;
  }

  switch (ENCODED_STATES(*this,s2))
  {
  case 0:  // this positive, s2 positive
    this->theUnion (*this, s2) ;
    break ;
  case 1:  // this positive, s2 negative
    s2.theDifference (*this, *this) ;
    this->U = true ;
    break ;
  case 2:  // this negative, s2 positive
    this->theDifference (*this, s2) ;
    break ;
  case 3:  // this negative, s2 negative
    this->theIntersection (*this, s2) ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for univeral set") ;
    break ;
  }
  
  // cleanup the input set operand and exit
  s2.cleanup() ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator-= (SetOf<T>& s2)
 *
 * Inputs:  'this' and 's2' are input sets, 
 *
 * Outputs: overwrites `this` with the set difference of `this` and 
 *          s2, i.e. `this` with all elements in s2 removed.
 *
 * Actions: first `this` and `s2` are threaded, and then scanned linearly
 *          to form the tree of the difference between s1 and s2.
 *
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::operator-= (SetOf<T>& s2)
{
  ++_oct_nb_differences ;
  CHECK_OPD(s2) ;

  // Axiom: this - {} = this

  if (!this || s2.is_empty())
    return ;
  
  // Axiom: {} - x = {}

  if (this->is_empty())
    return ;

  switch (ENCODED_STATES(*this,s2))
  {
  case 0:  // this positive, s2 positive */
    this->theDifference (*this, s2) ;
    break ;
  case 1:  // this positive, s2 negative */
    this->theIntersection (*this, s2) ;
    break ;
  case 2:  // this negative, s2 positive */
    this->theUnion (*this, s2) ;
    break ;
  case 3:  // this negative, s2 negative */
    s2.theDifference (*this, *this) ;
    this->U = false ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for complemented set") ;
    break ;
  }

  // clean up the input operand and exit
  s2.cleanup() ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator+= (int i) 
 *
 * Inputs:  'i' is a set member ID, and 'this' points to a set variable.
 *
 * Outputs: void.
 *
 * Actions: adds 'i' to set 'this' if !this->U, otherwise removes 'i'
 *          from 'this'. In the process it may modify the bounds of
 *          the Universe.
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::operator+= (int i) 
{
  int b = BASE_FUNCTION(i) ;
  
  ++_oct_nb_add_members ;

  if (u && i < u->l)
    u->l = i ;

  if (u && i > u->u)
    u->u = i ;

  SetNode *p = Map<SetNode>::operator[] (b) ;
  if (p)
  {
    if (this->U) {
      if (( (*p) -=  i-b )) {
	// the subset is completely empty, and its SetNode
	// should be unlinked from this set...
	this->delete_subset (p) ;
      }
    }
    else
      (*p) += i - b ; // add member i-b to SetNode **p
    return ;
  }
  
  // if we reach this point, then need to create a new node
  // provided s->U is false.

  if (!(this->U)) {
    SetNode *nr = new SetNode ;
    nr->base = b ;
    (*nr) += i - b ; // add member i-b to SetNode *nr
  
    // insert node 'nr' into the thread for 'this'
    
    this->insert_subset (nr) ;
    
    // install 'nr' in the hash table for 'this'
    
    this->install_subset (nr) ;
  }
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::operator-= (int i)
 *
 * Inputs:  'i' is a set element index, 'this' is a pointer to a set.
 *
 * Outputs: void.
 *
 * Actions: Afterwards, 'i' is not a member of 'this'. Thus, if U
 *          is false, we remove 'i' from its subset (if present),
 *          otherwise if U is true, we add 'i' to its subset
 *          or create such a subset if none already exists.
 *
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::operator-= (int i)
{
  int b = BASE_FUNCTION(i) ;
  
  ++_oct_nb_remove_members ;

  // Axiom: {} - {i} = {}
  // Axion: If i not a member of U, then Null effect

  if (!(this->thread) && !(this->U))
    return ;

  if (u && ((i < u->l) || (i > u->u)))
    return ;
  
  SetNode* p = Map<SetNode>::operator[] (b) ;
  if (p)
  {
    if (this->U)
      (*p) += i - b ; // add member i - b to SetNode **p
    else {
      if (( (*p) -= i - b )) {
	// subset *p is completely empty, and should be
	// deleted from this set...
	this->delete_subset (p) ;
      }
    }
    return ;
  }
  
  // If we reach this point, and this is is complemented set, then
  // we must create a new subset at base = b, and add (i-b) to it.
  
  if (this->U) {
    SetNode *nr = new SetNode ;
    nr->base = b ;
    (*nr) += i - b ;
  
    // insert node 'nr' into the thread for 'this'
    
    this->insert_subset (nr) ;
    
    // install 'nr' in the hash table for 'this'
    
    this->install_subset (nr) ;
  }
}


/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::operator[] (int i)
 *
 * Inputs:  's' is a set, and i is an integer ID for a member.
 *
 * Outputs: TRUE if {i} \subset `this`, else FALSE; i.e. tests for 
 *          element i being a member of set `this`
 * 
 * Actions: looks for element i in the set if !U, or for the absence
 *          of element i from the set if U.
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::operator[] (int i)
{
  ++_oct_nb_is_members ;

  // Axiom: s = {} => ~({i} \subset s)
  // ... in fact this axiom does not make the membership test
  //     more efficient, so we do not exploit it here.

  if (i < 0)
    FATAL_ERROR("set member index was negative")
  else {
    int b = BASE_FUNCTION(i) ;
    
    SetNode *p = Map<SetNode>::operator[] (b) ;
    if (p)
    {
      if (U)
	return (bool)!(*p)[i-b] ;
      else
	return (*p)[i-b] ;
    }
    return U ;
  }
  return false ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void Set::make_universal ()
 *
 * Inputs:  'this' is a pointer to a Set variable (itself a pointer to
 *               a set_node struct (or NULL).
 *
 * Outputs: always returns void, but creates a set containing all
 *          members having IDs from lb to ub inclusive.
 *
 * Actions: clears the set, and sets U to TRUE, and sets the bounds.
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::make_universal ()
{
  ++_oct_nb_universals ;

  CHECK_POINTER(this,"make_universal") ;

  if (!u) {
    FATAL_ERROR("A universal set must have a defined universe") ;
    return ;
  }

  this->empty() ;
  this->U = true ;
  this->vstamp = u ? u->vstamp : 0 ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::is_universal_set ()
 *
 * Inputs:  'this' is an input set.
 *
 * Outputs: true, if 'this' is the Universal set, but false otherwise.
 *
 * Actions: to be the universal set, the set must exist, its U bit
 *          must be set, and the list of absent members must be NULL.
 *
 *          A non-complemented set can also be the Universal
 *          set. To check for this we must do:-
 *
 *            s1 = U ;
 *            return (set_equivalence(s, s1)) ;
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::is_universal_set () 
{
  ++_oct_nb_universal_preds ;

  if (this) {
    if (U)
      return (bool)(!thread) ;
    else {
      if (!u) {
	FATAL_ERROR("To test for universality a set must have a defined universe") ;
	return false ;
      }
      SetOf<T> s (u) ;
      s.make_universal () ;
      return s == *this ;
    }
  }
  else
    return false ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: SetOf<T>& SetOf<T>::complement ()
 *
 * Inputs: 'this' is an input set
 *
 * Outputs: returns the bounded complement of `this`
 *
 * Actions: copys s1 to s0 and inverts the sense of the U bit.
 * 
 *-------------------------------------------------------------------*/

// TBD: CHECK UNARY OPERAND...

template <class T> SetOf<T>& SetOf<T>::complement ()
{
  // If this is not a temporary set we neet to create one, and note that
  // it is a temporary object (until assigned by Set::operator=)

  SetOf<T>* dst = this->temp ? this : new SetOf<T> (*this) ;
  SetOf<T>& s0 = *dst ;
  s0.temp = true ;
  s0.U = (bool)!(s0.U) ;  // invert sense of the U bit
  this->cleanup() ;       // in case SetOf<T> operand was temporary
  return s0 ;
}



/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::operator== (SetOf<T>& s2)
 *
 * Inputs:  `this` and s2 are two input sets.
 *
 * Outputs: true if contents of both sets are identical, else false.
 *
 * Actions: scan the thread list for s1 and s2, looking for any
 *          difference. 
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::operator== (SetOf<T>& s2)
{
  ++_oct_nb_equivalences ;
  CHECK_OPD(s2) ;
  
  // Try some simple checks first to catch set equality...

  // If both threads are the same, then if the threads are non-null,
  // or if both sets are of the same polarity, we know that they
  // are equivalent.
  
  if ((this->thread || (this->U == s2.U))
       && (this->thread == s2.thread))
    return true ;

  // ...else we deal with complementary set cases individually

  switch (ENCODED_STATES(*this,s2))
  {
  case 0:  // this positive, s2 positive
    return this->theEquivalence(s2) ;
    break ;
  case 1:  // this positive, s2 negative
    return this->fullEquivalence(s2) ;
    break ;
  case 2:  // this negative, s2 positive
    return s2.fullEquivalence(*this) ;
    break ;
  case 3:  // this negative, s2 negative
    return this->theEquivalence(s2) ;
    break ;
  default: // should never reach this point
    FATAL_ERROR ("illegal encoding for complemented set") ;
    break ;
  }
  return false ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::operator!= (SetOf<T> s2) 
 *
 * Inputs:  this and s2 are input sets, possibly empty.
 *
 * Outputs: TRUE if s1 \subset s2 and s2 \subset s1, else FALSE.
 *
 * Actions: simply resorts to using operator ==
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::operator!= (SetOf<T>& s2) 
{
  ++_oct_nb_non_equivalences ;

  if (!this)
    return false ;
  else
    return (bool)!(*this == s2) ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::is_empty()
 *
 * Inputs:  pointer to a set s
 *
 * Outputs: true if s is empty, false if s is non-empty
 *
 * Actions:
 *
 * Bugs: This will not work if a set becomes empty by virtue of
 *       having all its elements removed by the -= operator.
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::is_empty()
{
  ++_oct_nb_empty_preds ;

  if (!this)
    return true ;
  else if (!(this->thread))
    return ((bool)!(this->U)) ;
  else if (!(this->U))
    return false ;
  else{
    // this set has a thread, but may still be empty if the
    // contents of the complemented set covers the complete
    // universal domain for this set...
    for (int ii = this->lwb(); ii <= this->upb(); ii++)
      if ((*this)[ii])
	return false ;

    // If we reach this point, then the complemented set contains
    // all elements of the domain, and is therefore empty.
    // For space-efficiency, and later time-efficiency, we 
    // collapse this set into a regular empty set.

    this->empty() ;
    return true ;
  }
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::theIntersection (SetOf<T>& s0, SetOf<T>& s2)
 *
 * Inputs:  sets `this` and `s2`, and a pointer to result set s0.
 *
 * Outputs: void, but modifies *s0
 *
 * Actions: computes the intersection of the sets contained in the
 *          threads of `this` and `s2`, putting result in *s0.
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::theIntersection (SetOf<T>& s0, SetOf<T>& s2)
{
  SetNode *t1, *t2;
  SetNode **p0 ;

  // set up pointer-to-pointer to current result thread variable

  p0 = &(s0.thread) ;

  // scan the threads computing t0 = t1 \intersection t2

  t1 = this->thread ;
  t2 = s2.thread ;

  while (t1 && t2)
  {
    int i1 = t1 -> base ;
    int i2 = t2 -> base ;

    if (i1 == i2) {
      SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode(i1)) ;
      if (subset_intersection (nr->items, t1->items, t2->items)) {
	// subset intersection was non-empty, so include it
	nr->base = i1 ;            // so set base
	s0.install_subset(nr) ;    // install subset
	p0 = &((*p0)->thread) ;    // bump p0 pointer
      }
      // advance both input thread pointers
      t1 = t1 -> thread ;
      t2 = t2 -> thread ;
    }
    else if (i1 < i2) {
      // advance thread pointer for i1 only
      t1 = t1 -> thread ;
    }
    else /* i1 > i2 */ {
      // advance thread pointer for i2 only
      t2 = t2 -> thread ;
    }
  }
  
  // Now delete any remaining nodes in the thread for s0 
  // that were not used to store the result set.

  (*p0)->free_thread() ;
  *p0 = 0 ;

  return ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::theUnion (SetOf<T>& s0, SetOf<T>& s2)
 *
 * Inputs:  sets `this` and `s2`, and pointer to result set s0
 *
 * Outputs: void, but modifies *s0.
 *
 * Actions: performs the union operation on the threaded sets of
 *          `this` and `s2`, putting the result in *s0.
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::theUnion (SetOf<T>& s0, SetOf<T>& s2)
{
  SetNode *t1, *t2 ;
  SetNode **p0 ;

  // set up pointer-to-pointer to current s0 thread variable

  p0 = &(s0.thread) ;

  // scan the threads computing t0 = t1 U t2

  t1 = this->thread ;
  t2 = s2.thread ;

  while (t1 && t2)
  {
    int i1 = t1 -> base ;
    int i2 = t2 -> base ;

    if (i1 == i2) {
      SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode(i1)) ;
      subset_union (nr->items, t1->items, t2->items) ;
      nr->base = i1 ;
      s0.install_subset (nr) ;

      // advance all the thread pointers
      t1 = t1 -> thread ;
      t2 = t2 -> thread ;
      p0 = &((*p0)->thread) ;
    }
    else if (i1 < i2) {
      // advance thread pointer for i1, copying subset to s0
      // unless s0 is synonymous with s1.
      SetNode *nr = (*p0 && (*p0 != t2)) ? *p0 : new SetNode ;
      if (nr != *p0) {
	nr->thread = *p0 ;
	*p0 = nr ;
      }
      nr->base = t1->base ;
      copy_subset (t1->items, nr->items) ;
      s0.install_subset (nr) ;
      p0 = &((*p0)->thread) ;
      t1 = t1 -> thread ;
    }
    else /* i1 > i2 */ {
      // advance thread pointer for i2, copying to s0
      SetNode *nr = (*p0 && (*p0 != t1)) ? *p0 : new SetNode ;
      if (nr != *p0) {
	nr->thread = *p0 ;
	*p0 = nr ;
      }
      nr->base = t2->base ;
      copy_subset (t2->items, nr->items) ;
      s0.install_subset (nr) ;
      p0 = &((*p0)->thread) ;
      t2 = t2 -> thread ;
    }
  }

  // Now check if there are any remaining nodes in either t1 or t2,
  // as these need copying over to the output set.

  while (t1) {
    SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode) ;
    nr->base = t1->base ;
    if (t1->items != nr->items)
      copy_subset (t1->items, nr->items) ;
    s0.install_subset (nr) ;
    p0 = &((*p0)->thread) ;
    t1 = t1 -> thread ;
  }

  while (t2) {
    SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode) ;
    nr->base = t2->base ;
    if (t2->items != nr->items)
      copy_subset (t2->items, nr->items) ;
    s0.install_subset (nr) ;
    p0 = &((*p0)->thread) ;
    t2 = t2 -> thread ;
  }

  // Now delete any remaining nodes in the thread for s0 
  // that were not used to store the result set.

  (*p0)->free_thread() ;
  *p0 = 0 ;

  return ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: void SetOf<T>::theDifference (SetOf<T>& s0, SetOf<T>& s2)
 *
 * Inputs:  sets `this` and `s2` and pointer to result set s0.
 *
 * Outputs: void, but modifies *s0.
 *
 * Actions: computes the set `this` with elements of `s2` removed, and
 *          assigns this to &s0.
 * 
 *-------------------------------------------------------------------*/

template <class T> void SetOf<T>::theDifference (SetOf<T>& s0, SetOf<T>& s2)
{
  SetNode *t1, *t2;
  SetNode **p0 ;

  // Set up pointer-to-pointer to current s0 thread variable.

  p0 = &(s0.thread) ;

  // scan the threads computing t0 = t1 - t2

  t1 = this->thread ;
  t2 = s2.thread ;

  while (t1 && t2)
  {
    int i1 = t1 -> base ;
    int i2 = t2 -> base ;

    if (i1 == i2) {
      SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode) ;
      if (subset_difference (nr->items, t1->items, t2->items)) {
	// result was not empty
	nr->base = i1 ;               // so set base
	s0.install_subset (nr) ;     // install, 
	p0 = &((*p0)->thread) ;       // and bump p0 pointer
      }
      // advance both input thread pointers
      t1 = t1 -> thread ;
      t2 = t2 -> thread ;
    }
    else if (i1 < i2) {
      // advance thread pointer for i1, copying it over to s0
      SetNode *nr = (*p0 && (*p0 != t2)) ? *p0 : new SetNode ;
      if (nr != *p0) {
	nr->thread = *p0 ;
	*p0 = nr ;
      }
      nr->base = t1->base ;
      copy_subset (t1->items, nr->items) ;
      s0.install_subset (nr) ;
      p0 = &((*p0)->thread) ;
      t1 = t1 -> thread ;
    }
    else /* i1 > i2 */ {
      // advance thread pointer for i2
      t2 = t2 -> thread ;
    }
  }

  // Now check if there are any remaining nodes in t1,
  // as these need copying over to the output set.

  while (t1) {
    SetNode *nr = *p0 ? *p0 : (*p0 = new SetNode) ;
    nr->base = t1->base ;
    if (t1->items != nr->items)
      copy_subset (t1->items, nr->items) ;
    s0.install_subset (nr) ;
    p0 = &((*p0)->thread) ;
    t1 = t1 -> thread ;
  }

  // Now delete any remaining nodes in the thread for s0 
  // that were not used to store the result set.

  (*p0)->free_thread() ;
  *p0 = 0 ;

  return ;
}



/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::theEquivalence (SetOf<T>& s2)
 *
 * Inputs:  sets `this` and `s2`
 *
 * Outputs: checks if the threaded contents of `this` and `s2` are
 *          the same.
 *
 * Actions: scans the threads of `this` and `s2`, calling 
 *          subset_equivalence()
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::theEquivalence (SetOf<T>& s2)
{
  SetNode *t1, *t2 ;

  // Resort to scanning the sets
  
  for (t1 = this->thread, t2 = s2.thread; 
       t1 && t2; 
       t1 = t1->thread, t2 = t2->thread)
  {
    if ((t1->base != t2->base) || 
	!subset_equivalence(t1->items, t2->items))
      return false ;
  }
  if (t1 || t2)
    return false ;
  else
    return true ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: bool SetOf<T>::fullEquivalence (SetOf<T>& s2)
 *
 * Inputs:  Sets s1 and s2.
 *
 * Outputs: The equivalence of `this` and (U-s2).
 *
 * Actions: Scan the threads of `this` and `s2` to see if :-
 * 
 *     (i). The intersections of common subsets of `this` and `s2` are
 *          always empty.
 *
 *    (ii). The union of all subsets in `this` and `s2` covers completely 
 *          the universal range defined by u->lwb thru u->upb
 *
 *  If either of these two conditions fails, then `this` and (U-s2) are
 *  not equivalent.
 * 
 *-------------------------------------------------------------------*/

template <class T> bool SetOf<T>::fullEquivalence (SetOf<T>& s2)
{
  subSet maskset ;
  SetNode *t1 = this->thread ;
  SetNode *t2 = s2.thread ;
  int lb = u->lwb() ;
  int ub = u->upb() ;
  int topBase = BASE_FUNCTION(ub) ;
  int bottomBase = BASE_FUNCTION(lb) ;
  int topElem = ub - topBase ;
  int bottomElem = lb - bottomBase ;
  int currentBase = topBase ;
  int more_subsets = true ;
  SetNode *firstSet = 0 ;
  int i1 = -1, i2 = -1;
  int mmin, mmax ;

  // Find the subset with the highest base index

  if (t1) {
    i1 = t1->base ;
    if (t2) {
      i2 = t2->base ;
      firstSet = (i1 < i2) ? t1 : t2 ;
    }
    else
      firstSet = t1 ;
  }
  else {
    firstSet = t2 ;
    if (t2)
      i2 = t2->base ;
  }
    
  if (!firstSet)
    return false ;

  while (more_subsets)
  {
    // If we are not finished checking, and both input lists are
    // exhausted, we must return FALSE.

    if (!t1 && !t2)
      return false ;

    // Is this the final subset to consider ?

    more_subsets = currentBase != bottomBase ;

    // Set up boundary conditions for this subset comparison

    mmin = (currentBase == bottomBase) ? bottomElem : 0 ;
    mmax = (currentBase == topBase) ? topElem : CARDINALITY - 1 ;

    // Create a mask set for the subset elements of interest

    make_subset (maskset, mmin, mmax) ;

    // Extract the base indices, i1 and i2, for the current
    // subsets of t1 and t2.

    if (t1)
      i1 = t1->base ;
    else
      i1 = lb - 1 ;
    if (t2)
      i2 = t2->base ;
    else
      i2 = lb - 1 ;

    // Compare the current subsets of t1 and t2

    if (i1 == i2) {
      SetNode tmp, eset ;
      if (i1 != currentBase)
	return false ;

      subset_union (tmp.items, 
		    t1 ? t1->items : eset.items,
		    t2 ? t2->items : eset.items) ;

      if (!subset_equivalence(tmp.items, maskset))
	return false ;

      subset_intersection (tmp.items, 
			   t1 ? t1->items : eset.items, 
			   t2 ? t2->items : eset.items) ;

      if (!tmp.is_empty_subset())
	return false ;

      currentBase = currentBase - CARDINALITY ;
      t1 = t1 ? t1->thread : 0 ;
      t2 = t2 ? t2->thread : 0 ;
    }
    if (i1 < i2)
    {
      SetNode eset ;
      if (i1 != currentBase)
	return false ;
      if (!subset_equivalence (t1 ? t1->items : eset.items, maskset))
	return false ;
      currentBase = currentBase - CARDINALITY ;
      t1 = t1 ? t1->thread : 0 ;
    }
    if (i2 < i1) 
    {
      SetNode eset ;;;
      if (i2 != currentBase)
	return false ;
      if (!subset_equivalence (t2 ? t2->items : eset.items, maskset))
	return false ;
      currentBase = currentBase - CARDINALITY ;
      t2 = t2 ? t2->thread : 0 ;
    }
  }

  // If we terminate the 'while (more_subsets)' loop without
  // returning FALSE, then we can assert that `this` is the complement
  // of `s2`, and (this U s2) equals the Universal set bounded by
  // (u->l, u->u).

  return true ;
}


/*--------------------------------------------------------------------
 * 
 * Routine: char* SetOf<T>::print (int m, FILE* f)
 *
 * Inputs:  'm' is an integer flag
 *
 *          'f' is a FILE pointer to send output to if m is false.
 *
 * Outputs: if m is true, it returns a pointer to a newly allocated
 *          byte string containing the printed string.
 *
 * Actions: iterates over Set `this`,  printing the set in a 
 *          complemented format.
 * 
 *-------------------------------------------------------------------*/

template <class T> char* SetOf<T>::print (int m, FILE* f)
{
  int slen = 0 ;
  char *s = NULL, *s2 = NULL ;
  int lb = u ? u->l : 0 ;
  int ub = u ? u->u : 0 ;

  if (this && this->U) {
    if (m) {
      s = new char[MAX_SET_STRING] ;
      s2 = new char[ELEMENT_STRING_SIZE] ;
      sprintf (s, "U(%d,%d) - {", lb, ub) ;
      slen += strlen (s) ;
    }
    else {
      fprintf (f, "U(%d,%d) - {", lb, ub) ;
    }
  }
  else {
    if (m) {
      s = new char[MAX_SET_STRING] ;
      s2 = new char[ELEMENT_STRING_SIZE] ;
      sprintf (s, "{") ;
      slen += strlen (s) ;
    }
    else
      fprintf (f, "{") ;
  }

  SetIteratorOf<T> next (*this, SET_STORED) ;
  for (int ii, comma = false; ii = next(); ++next)
  {
    if (m) {
      if (comma)
	sprintf (s2, ", %d", ii) ;
      else
	sprintf (s2, "%d", ii) ;
      int s2l = strlen (s2) ;
      if ( (slen + s2l + 10) > MAX_SET_STRING ) {
	// can't print the whole set...its too big
	sprintf (s2, ", ... ") ;
	s = strcat (s, s2) ;
	slen += 6 ;
	break ;
      }
      else {
	s = strcat (s, s2) ;
	slen += s2l ;
      }
    }
    else {
      if (comma)
	fprintf (f, ", %d", ii) ;
      else
	fprintf (f, "%d", ii) ;
    }
    comma = true ;
  }
  
  if (m) {
    sprintf (s2, "}") ;
    s = strcat (s, s2) ;
  }
  else
    fprintf (f, "}") ;

  delete s2 ;

  if (m)
    return s ;
  else
    return NULL ;
}

#endif // _sets_h_
